from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import math

# {{{ symmetry operations


class SymmetryOperationBase(object):
    def __init__(self, index):
        self._index = index

    def __lt__(self, other):
        if type(self) == type(other):
            return self._index < other._index

        # differnt operations in lexicographical order
        return repr(self) < repr(other)


class Flip(SymmetryOperationBase):
    """
    Flip the sign of an axis, spanning S_2^dim
    """

    def __init__(self, iaxis):
        super(Flip, self).__init__(iaxis)
        self.axis = iaxis

    def __repr__(self):
        return "Flip(%d)" % self.axis


class Swap(SymmetryOperationBase):
    """
    Swap two axes, spanning S_dim
    """

    def __init__(self, iaxis, jaxis):
        self.axes = (iaxis, jaxis)
        super(Swap, self).__init__(sorted(self.axes))

    def __repr__(self):
        return "Swap(%d,%d)" % tuple(sorted(self.axes))


# }}} End symmetry operations


class CaseVecReduction(object):
    """
    Reduce a set of case vectors based on symmetry.
    """

    def __init__(self, vecs=None, sym_tags=None, do_reduction=True):
        """
        sym_tags is a list of SymmetryOperationBase objects.
        sym_tags is [] if no symmetry can be used.
        sym_tags is None if maximum symmetry can be used.
        """
        if isinstance(vecs, list):
            assert len(vecs) >= 1
        else:
            raise RuntimeError("Invalid list of case vecs.")

        self.dim = len(vecs[0])
        for vec in vecs:
            assert len(vec) == self.dim
        self.full_vecs = vecs

        if sym_tags is not None:
            for tag in sym_tags:
                assert isinstance(tag, SymmetryOperationBase)
        self.symmetry_tags = sym_tags
        self.flippable, self.swappable_groups = self.parse_symmetry_tags(
            self.symmetry_tags
        )

        self.reduced = False
        if do_reduction:
            self.reduce()

    def parse_symmetry_tags(self, tags):
        flippable = np.zeros(self.dim)
        swappable_groups = []

        if tags is None:
            flippable += 1
            swappable_groups.append({i for i in range(self.dim)})
            return flippable, swappable_groups

        for tag in tags:

            if isinstance(tag, Flip):
                flippable[tag.axis] = 1

            elif isinstance(tag, Swap):
                iaxis, jaxis = tag.axes
                gi = None
                gj = None
                for gid, group in zip(
                        range(len(swappable_groups)),
                        swappable_groups
                        ):
                    if iaxis in group:
                        assert gi is None
                        gi = gid
                    if jaxis in group:
                        assert gj is None
                        gj = gid
                if gi is None:
                    if gj is None:
                        # New group
                        swappable_groups.append({iaxis, jaxis})
                    else:
                        # Update group[gj]
                        swappable_groups[gj].add(iaxis)
                else:
                    if gj is None:
                        # Update group[gi]
                        swappable_groups[gi].add(jaxis)
                    else:
                        if gi == gj:
                            pass
                        else:
                            # Merge groups
                            swappable_groups.append(
                                set().union(
                                    swappable_groups[gi],
                                    swappable_groups[gj])
                                )
                            swappable_groups.remove(swappable_groups[gi])
                            swappable_groups.remove(swappable_groups[gj])

            else:
                raise NotImplementedError

        return flippable, swappable_groups

    def find_base_vecs(self):
        vecs = self.full_vecs
        base_vecs = []
        base_vec_ids = []
        for vid, vec in zip(range(len(vecs)), vecs):
            is_base = True
            # Check for flips
            for d in range(self.dim):
                v = vec[d]
                if not self.flippable[d]:
                    continue
                else:
                    if v > 0:
                        is_base = False
                        break
            # Check for swaps
            if is_base:
                for group in self.swappable_groups:
                    group_view = list(group)
                    vec_part = np.array(vec)[group_view]
                    if sorted(vec_part) != list(vec_part):
                        is_base = False
                        break
            if is_base:
                base_vecs.append(vec)
                base_vec_ids.append(vid)
        return base_vecs, base_vec_ids

    def find_invariant_group(self, vec):
        """
        For a given case vector, within the allowed symmetry tags,
        return a generating set of its invariant group as a list of
        SymmetryOperationBase objects.
        """
        ivgp = []
        n = len(vec)
        assert n == self.dim
        for iaxis in range(n):
            if not self.flippable[iaxis]:
                continue
            if vec[iaxis] == -vec[iaxis]:
                ivgp.append(Flip(iaxis))
        for iaxis in range(n):
            for jaxis in range(iaxis + 1, n):
                if vec[iaxis] == vec[jaxis]:
                    # only if swap(i,j) is allowed
                    for group in self.swappable_groups:
                        if (iaxis in group) and (jaxis in group):
                            ivgp.append(Swap(iaxis, jaxis))
                            break
        return ivgp

    def reduce(self):
        self.reduced_vecs, self.reduced_vec_ids = self.find_base_vecs()
        self.reduced_invariant_groups = [
            self.find_invariant_group(v) for v in self.reduced_vecs
        ]
        self.reduced = True

    # call reduce() before calling getters

    def get_reduced_vecs(self):
        assert self.reduced
        return self.reduced_vecs

    def get_reduced_vec_ids(self):
        assert self.reduced
        return self.reduced_vec_ids

    def get_inter_box_reduction_ratio(self):
        assert self.reduced
        return len(self.reduced_vecs) / len(self.full_vecs)

    def get_intra_box_reduction_ratio(self):
        assert self.reduced
        total_ratio = 0
        for vid in range(len(self.reduced_vecs)):
            ratio = 1
            fable, sgroups = self.parse_symmetry_tags(
                self.reduced_invariant_groups[vid]
            )
            ratio = ratio / (2 ** (sum(fable)))
            for grp in sgroups:
                ratio = ratio / math.factorial(len(grp))
            total_ratio += ratio
        return total_ratio / len(self.reduced_vecs)

    def get_full_reduction_ratio(self):
        return (
            self.get_inter_box_reduction_ratio()
            * self.get_intra_box_reduction_ratio()
        )

    def get_reduced_invariant_groups(self):
        assert self.reduced
        return self.reduced_invariant_groups



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2019 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import loopy as lp
import pymbolic as pmbl
from volumential.tools import ScalarFieldExpressionEvaluation

# {{{ math functions

CL_MATH_URL = r"https://www.khronos.org/registry/OpenCL/sdk/1.0/docs/man/xhtml/mathFunctions.html"  # noqa

CL_MATH_FUNCS = [
        'acos',     'acosh',     'acospi',  'asin',
        'asinh',    'asinpi',    'atan',    'atan2',
        'atanh',    'atanpi',    'atan2pi', 'cbrt',
        'ceil',     'copysign',  'cos',     'cosh',
        'cospi',    'erfc',      'erf',     'exp',
        'exp2',     'exp10',     'expm1',   'fabs',
        'fdim',     'floor',     'fma',     'fmax',
        'fmin',     'fmod',      'fract',   'frexp',
        'hypot',    'ilogb',     'ldexp',   'lgamma',
        'lgamma_r', 'log',       'log2',    'log10',
        'log1p',    'logb',      'mad',     'modf',
        'nan',      'nextafter', 'pow',     'pown',
        'powr',     'remainder', 'remquo',  'rint',
        'rootn',    'round',     'rsqrt',   'sin',
        'sincos',   'sinh',      'sinpi',   'sqrt',
        'tan',      'tanh',      'tanpi',   'tgamma',
        'trunc',
        ]

clmath_decl_code = r"""
def FUNC_NAME(x):
    "CL math function FUNC_NAME.\n\nSee CL_MATH_URL for details."
    return pmbl.primitives.Call(
            pmbl.primitives.Lookup(pmbl.primitives.Variable("math"), "FUNC_NAME"),
            (x,))
"""

for fname in CL_MATH_FUNCS:
    code = clmath_decl_code.replace(
            'FUNC_NAME', fname).replace(
                    'CL_MATH_URL', CL_MATH_URL)
    exec(code)

# }}} End math functions

x = pmbl.var("x")
y = pmbl.var("y")
z = pmbl.var("z")


def der_laplacian(func, coord_vars=["x", "y", "z"]):
    return sum(pmbl.diff(pmbl.diff(func, var), var)
            for var in coord_vars)

# {{{ evaluation helper


def math_func_mangler(target, name, arg_dtypes):
    """Magic function that is necessary for evaluating math functions
    """
    if len(arg_dtypes) == 1 and isinstance(name, pmbl.primitives.Lookup):
        arg_dtype, = arg_dtypes

        fname = name.name
        if not (isinstance(name.aggregate, pmbl.primitives.Variable)
                and name.aggregate.name == 'math'):
            raise RuntimeError("unexpected aggregate '%s'" %
                    str(name.aggregate))

        if arg_dtype.is_complex():
            if arg_dtype.numpy_dtype == np.complex64:
                tpname = "cfloat"
            elif arg_dtype.numpy_dtype == np.complex128:
                tpname = "cdouble"
            else:
                raise RuntimeError("unexpected complex type '%s'" %
                        arg_dtype)

            return lp.CallMangleInfo(
                   target_name="%s_%s" % (tpname, fname),
                   result_dtypes=(arg_dtype,),
                   arg_dtypes=(arg_dtype,))

        else:
            return lp.CallMangleInfo(
                   target_name="%s" % fname,
                   result_dtypes=(arg_dtype,),
                   arg_dtypes=(arg_dtype,))

    return None


def get_evaluator(dim, expression, variables=None):
    if variables is None:
        if dim == 1:
            variables = [x]
        elif dim == 2:
            variables = [x, y]
        elif dim == 3:
            variables = [x, y, z]
    else:
        assert len(variables) == dim

    return ScalarFieldExpressionEvaluation(
            dim=dim,
            expression=expression,
            variables=variables,
            function_manglers=[math_func_mangler])

# }}} End evaluation helper



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import volumential.nearfield_potential_table as npt
from sumpy.point_calculus import CalculusPatch


dim = 2
patch_order = 6

s = 1.5
epsilon = 0.01
delta_t = 0.05
final_t = delta_t * 1
theta_y = 60.0 / 180.0 * np.pi

b = s / (epsilon ** 2)
c = 1.0 / (epsilon * delta_t)


def make_patch(center, size):
    return CalculusPatch(center=center, h=size, order=patch_order)


def test_patch_cahn_hilliard():

    rep = 3
    size = 0.05

    import random

    for r in range(rep):
        center_x = random.uniform(-1, 1)
        center_y = random.uniform(-1, 1)
        center = [center_x, center_y]
        patch = make_patch(center, size)

        f_values = 0.5 * 1 / (4 * 3 * 2) * (patch.x ** 4 + patch.y ** 4)

        lap = patch.laplace(f_values)
        laplap = patch.laplace(lap)

        # print(center, size, len(patch.x), lap, laplap)

        assert np.max(np.abs(laplap - 1)) < 5e-6


def test_kernel_func():

    # With/without special expansion at origin
    knl_func = npt.get_cahn_hilliard(dim, b, c)
    alternate_knl_func = npt.get_cahn_hilliard(dim, b, c, True)

    for i in range(50):
        ri = 0.001 * 2 ** (-i)
        assert abs(knl_func(ri, 0) - alternate_knl_func(ri, 0)) < 1e-12


def direct_quad(source_func, target_point):

    knl_func = npt.get_cahn_hilliard(dim, b, c)

    def integrand(x, y):
        return source_func(x, y) * knl_func(target_point[0] - x, target_point[1] - y)

    import volumential.singular_integral_2d as squad

    integral, error = squad.box_quad(
        func=integrand, a=0, b=1, c=0, d=1, singular_point=target_point, maxiter=100
    )

    return integral


def eval_f(patch, source_func=None):
    n = len(patch.x)
    ff = np.zeros(n)

    def const_source(x, y):
        return 1

    if source_func is None:
        source_func = const_source

    for i in range(n):
        ff[i] = direct_quad(source_func, (patch.x[i], patch.y[i]))

    return ff


def test_cahn_hilliard_same_box_on_patch(longrun):

    rep = 1
    # Prevent numerical instability when point spacings are too small
    size = 0.5

    # inside the box
    import random

    for r in range(rep):
        center_x = random.uniform(size / 2, 1 - size / 2)
        center_y = random.uniform(size / 2, 1 - size / 2)
        center = [center_x, center_y]
        patch = make_patch(center, size)

        def source_func(x, y):
            return c

        f_values = eval_f(patch, source_func)

        lap = patch.laplace(f_values)
        laplap = patch.laplace(lap)

        # print(c)
        # print(center)
        # print((laplap - b * lap + c * f_values))
        # print((laplap - b * lap + c * f_values - c) / c)

        assert np.max(np.abs(laplap - b * lap + c * f_values - c) / c) < 5e-2


# def test_cahn_hilliard_table():
# assert False


if __name__ == "__main__":
    import sys

    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from py.test.cmdline import main

        main([__file__])



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__doc__ = """
.. autofunction:: drive_volume_fmm
.. autofunction:: interpolate_volume_potential
"""

import logging

logger = logging.getLogger(__name__)

import pyopencl as cl
from boxtree.fmm import TimingRecorder
from volumential.expansion_wrangler_interface import ExpansionWranglerInterface
from volumential.expansion_wrangler_fpnd import (
        FPNDSumpyExpansionWrangler, FPNDFMMLibExpansionWrangler)

import numpy as np


def drive_volume_fmm(
    traversal,
    expansion_wrangler,
    src_weights,
    src_func,
    direct_evaluation=False,
    timing_data=None,
    **kwargs
):
    """
    Top-level driver routine for volume potential calculation
    via fast multiple method.

    This function, and the interface it utilizes, is adapted from boxtree/fmm.py

    The fast multipole method is a two-pass algorithm:

    1. During the fist (upward) pass, the multipole expansions for all boxes
    at all levels are formed from bottom up.

    2. In the second (downward) pass, the local expansions for all boxes
    at all levels at formed from top down.

    :arg traversal: A :class:`boxtree.traversal.FMMTraversalInfo` instance.
    :arg expansion_wrangler: An object implementing the expansion
                    wrangler interface.
    :arg src_weights: Source 'density/weights/charges' time quad weights..
        Passed unmodified to *expansion_wrangler*.
    :arg src_func: Source 'density/weights/charges' function.
        Passed unmodified to *expansion_wrangler*.

    Returns the potentials computed by *expansion_wrangler*.
    """
    wrangler = expansion_wrangler
    assert issubclass(type(wrangler), ExpansionWranglerInterface)

    recorder = TimingRecorder()
    logger.info("start fmm")

    if isinstance(expansion_wrangler, FPNDSumpyExpansionWrangler):
        assert isinstance(src_weights, cl.array.Array)
        assert isinstance(src_func, cl.array.Array)

    elif isinstance(expansion_wrangler, FPNDFMMLibExpansionWrangler):

        traversal = traversal.get(wrangler.queue)

        if isinstance(src_weights, cl.array.Array):
            src_weights = src_weights.get(wrangler.queue)
        if isinstance(src_func, cl.array.Array):
            src_func = src_func.get(wrangler.queue)

    logger.debug("reorder source weights")

    src_weights = wrangler.reorder_sources(src_weights)
    src_func = wrangler.reorder_targets(src_func)

    # {{{ Construct local multipoles

    logger.debug("construct local multipoles")
    mpole_exps, timing_future = wrangler.form_multipoles(
        traversal.level_start_source_box_nrs, traversal.source_boxes, src_weights
    )
    recorder.add("form_multipoles", timing_future)

    # print(max(abs(mpole_exps)))

    # }}}

    # {{{ Propagate multipoles upward

    logger.debug("propagate multipoles upward")
    mpole_exps, timing_future = wrangler.coarsen_multipoles(
        traversal.level_start_source_parent_box_nrs,
        traversal.source_parent_boxes,
        mpole_exps,
    )
    recorder.add("coarsen_multipoles", timing_future)

    # print(max(abs(mpole_exps)))

    # mpole_exps is called Phi in [1]

    # }}}

    # {{{ Direct evaluation from neighbor source boxes ("list 1")

    logger.debug("direct evaluation from neighbor source boxes ('list 1')")
    # look up in the prebuilt table
    # this step also constructs the output array
    potentials, timing_future = wrangler.eval_direct(
        traversal.target_boxes,
        traversal.neighbor_source_boxes_starts,
        traversal.neighbor_source_boxes_lists,
        src_func,
    )
    recorder.add("eval_direct", timing_future)

    # Return list 1 only, for debugging
    # 'list1_only' takes precedence over 'exclude_list1'
    if 'list1_only' in kwargs and kwargs['list1_only']:

        logger.debug("reorder potentials")
        result = wrangler.reorder_potentials(potentials)

        logger.debug("finalize potentials")
        result = wrangler.finalize_potentials(result)

        logger.info("fmm complete with list 1 only")
        logger.info("fmm complete")
        logger.warn("only list 1 results are returned")

        return result

    # Do not include list 1
    if 'exclude_list1' in kwargs and kwargs['exclude_list1']:
        logger.info("Using zeros for list 1")
        logger.warn("list 1 interactions are not included")
        potentials = wrangler.output_zeros()

    # these potentials are called alpha in [1]

    # }}}

    # {{{ (Optional) direct evaluation of everything and return
    if direct_evaluation:

        print("Warning: NOT USING FMM (forcing global p2p)")
        if len(src_weights) != len(src_func):
            print("Using P2P with different src/tgt discretizations can be "
                  "unstable when targets are close to the sources while not "
                  "be exactly the same")

        # list 2 and beyond
        # First call global p2p, then substract list 1

        from sumpy import P2P

        dtype = wrangler.dtype

        p2p = P2P(
            wrangler.queue.context,
            wrangler.code.out_kernels,
            exclude_self=wrangler.code.exclude_self,
            value_dtypes=[wrangler.dtype],
        )

        p2p_extra_kwargs = dict(wrangler.self_extra_kwargs)
        p2p_extra_kwargs.update(wrangler.kernel_extra_kwargs)

        evt, (ref_pot,) = p2p(
            wrangler.queue,
            traversal.tree.targets,
            traversal.tree.sources,
            (src_weights.astype(dtype),),
            **p2p_extra_kwargs
        )
        potentials[0] += ref_pot

        l1_potentials, timing_future = wrangler.eval_direct_p2p(
                traversal.target_boxes,
                traversal.neighbor_source_boxes_starts,
                traversal.neighbor_source_boxes_lists,
                src_weights,
                )

        potentials = potentials - l1_potentials

        # list 3
        assert traversal.from_sep_close_smaller_starts is None

        # list 4
        assert traversal.from_sep_close_bigger_starts is None

        logger.debug("reorder potentials")
        result = wrangler.reorder_potentials(potentials)

        logger.debug("finalize potentials")
        result = wrangler.finalize_potentials(result)

        logger.info("direct p2p complete")

        return result

    # }}} End Stage

    # {{{ Translate separated siblings' ("list 2") mpoles to local

    logger.debug("translate separated siblings' ('list 2') mpoles to local")
    local_exps, timing_future = wrangler.multipole_to_local(
        traversal.level_start_target_or_target_parent_box_nrs,
        traversal.target_or_target_parent_boxes,
        traversal.from_sep_siblings_starts,
        traversal.from_sep_siblings_lists,
        mpole_exps,
    )
    recorder.add("multipole_to_local", timing_future)

    # print(max(abs(local_exps)))

    # local_exps represents both Gamma and Delta in [1]

    # }}}

    # {{{ Evaluate sep. smaller mpoles ("list 3") at particles

    logger.debug("evaluate sep. smaller mpoles at particles ('list 3 far')")

    # (the point of aiming this stage at particles is specifically to keep its
    # contribution *out* of the downward-propagating local expansions)

    mpole_result, timing_future = wrangler.eval_multipoles(
        traversal.target_boxes_sep_smaller_by_source_level,
        traversal.from_sep_smaller_by_level,
        mpole_exps,
    )
    recorder.add("eval_multipoles", timing_future)

    potentials = potentials + mpole_result

    # these potentials are called beta in [1]

    # volume fmm does not work with list 3 close currently
    # but list 3 should be empty with our use cases
    assert traversal.from_sep_close_smaller_starts is None
    # if traversal.from_sep_close_smaller_starts is not None:
    #    logger.debug("evaluate separated close smaller interactions directly "
    #                 "('list 3 close')")

    #    potentials = potentials + wrangler.eval_direct(
    #        traversal.target_boxes, traversal.from_sep_close_smaller_starts,
    #        traversal.from_sep_close_smaller_lists, src_weights)

    # }}}

    # {{{ Form locals for separated bigger source boxes ("list 4")

    logger.debug("form locals for separated bigger source boxes ('list 4 far')")

    local_result, timing_future = wrangler.form_locals(
        traversal.level_start_target_or_target_parent_box_nrs,
        traversal.target_or_target_parent_boxes,
        traversal.from_sep_bigger_starts,
        traversal.from_sep_bigger_lists,
        src_weights,
    )

    recorder.add("form_locals", timing_future)

    local_exps = local_exps + local_result

    # volume fmm does not work with list 4 currently
    assert traversal.from_sep_close_bigger_starts is None
    # if traversal.from_sep_close_bigger_starts is not None:
    #    logger.debug("evaluate separated close bigger interactions directly "
    #                 "('list 4 close')")

    #    potentials = potentials + wrangler.eval_direct(
    #        traversal.target_or_target_parent_boxes,
    #        traversal.from_sep_close_bigger_starts,
    #        traversal.from_sep_close_bigger_lists, src_weights)

    # }}}

    # {{{ Propagate local_exps downward

    logger.debug("propagate local_exps downward")
    # import numpy.linalg as la

    local_exps, timing_future = wrangler.refine_locals(
        traversal.level_start_target_or_target_parent_box_nrs,
        traversal.target_or_target_parent_boxes,
        local_exps,
    )

    recorder.add("refine_locals", timing_future)

    # }}}

    # {{{ Evaluate locals

    logger.debug("evaluate locals")

    local_result, timing_future = wrangler.eval_locals(
        traversal.level_start_target_box_nrs, traversal.target_boxes, local_exps
    )

    recorder.add("eval_locals", timing_future)

    potentials = potentials + local_result

    # }}}

    # {{{ Reorder potentials
    logger.debug("reorder potentials")
    result = wrangler.reorder_potentials(potentials)

    logger.debug("finalize potentials")
    result = wrangler.finalize_potentials(result)

    logger.info("fmm complete")
    # }}} End Reorder potentials

    if timing_data is not None:
        timing_data.update(recorder.summarize())

    return result


def compute_barycentric_lagrange_params(q_order):

    # 1d quad points and weights
    q_points_1d, q_weights_1d = np.polynomial.legendre.leggauss(q_order)
    q_points_1d = (q_points_1d + 1) * 0.5
    q_weights_1d *= 0.5

    # interpolation weights for barycentric Lagrange interpolation
    from scipy.interpolate import BarycentricInterpolator as Interpolator

    interp = Interpolator(xi=q_points_1d, yi=None)
    interp_weights = interp.wi
    interp_points = interp.xi

    return (interp_points, interp_weights)


def interpolate_volume_potential(
    target_points, traversal, wrangler, potential, target_radii=None, lbl_lookup=None
):
    """
    Interpolate the volume potential, only works for tensor-product quadrature formulae.
    target_points and potential should be an cl array.

    wrangler is used only for general info (nothing sumpy kernel specific)
    """

    dim = next(iter(wrangler.near_field_table.values()))[0].dim
    tree = wrangler.tree
    queue = wrangler.queue
    q_order = wrangler.quad_order
    ctx = queue.context
    dtype = wrangler.dtype
    coord_dtype = tree.coord_dtype
    n_points = len(target_points[0])

    from volumential.expansion_wrangler_fpnd import FPNDFMMLibExpansionWrangler
    if isinstance(wrangler, FPNDFMMLibExpansionWrangler):
        tree = tree.to_device(queue)
        traversal = traversal.to_device(queue)

    assert dim == len(target_points)

    # Building the lookup takes O(n*log(n))
    if lbl_lookup is None:
        from boxtree.area_query import LeavesToBallsLookupBuilder

        lookup_builder = LeavesToBallsLookupBuilder(ctx)

        if target_radii is None:
            import pyopencl as cl

            # Set this number small enough so that all points found
            # are inside the box
            target_radii = cl.array.to_device(
                queue, np.ones(n_points, dtype=coord_dtype) * 1e-12
            )

        lbl_lookup, evt = lookup_builder(queue, tree, target_points, target_radii)

    pout = cl.array.zeros(queue, n_points, dtype=dtype)
    multiplicity = cl.array.zeros(queue, n_points, dtype=dtype)
    pout.add_event(evt)

    # map all boxes to a template [0,1]^2 so that the interpolation
    # weights and modes can be precomputed
    (blpoints, blweights) = compute_barycentric_lagrange_params(q_order)
    blpoints = cl.array.to_device(queue, blpoints)
    blweights = cl.array.to_device(queue, blweights)

    # {{{ loopy kernel for interpolation

    if dim == 1:
        code_target_coords_assignment = """target_coords_x[target_point_id]"""
    if dim == 2:
        code_target_coords_assignment = """if(
        iaxis == 0, target_coords_x[target_point_id],
                    target_coords_y[target_point_id])"""
    elif dim == 3:
        code_target_coords_assignment = """if(
        iaxis == 0, target_coords_x[target_point_id], if(
        iaxis == 1, target_coords_y[target_point_id],
                    target_coords_z[target_point_id]))"""
    else:
        raise NotImplementedError

    if dim == 1:
        code_mode_index_assignment = """mid"""
    elif dim == 2:
        code_mode_index_assignment = """if(
        iaxis == 0, mid / Q_ORDER,
                    mid % Q_ORDER)""".replace(
            "Q_ORDER", "q_order"
        )
    elif dim == 3:
        code_mode_index_assignment = """if(
        iaxis == 0, mid / (Q_ORDER * Q_ORDER), if(
        iaxis == 1, mid % (Q_ORDER * Q_ORDER) / Q_ORDER,
                    mid % (Q_ORDER * Q_ORDER) % Q_ORDER))""".replace(
            "Q_ORDER", "q_order"
        )
    else:
        raise NotImplementedError

    import loopy

    # FIXME: noqa for specific lines does not work here
    # flake8: noqa
    lpknl = loopy.make_kernel(  # noqa
        [
            "{ [ tbox, iaxis ] : 0 <= tbox < n_tgt_boxes " "and 0 <= iaxis < dim }",
            "{ [ tpt, mid, mjd, mkd ] : tpt_begin <= tpt < tpt_end "
            "and 0 <= mid < n_box_modes and 0 <= mjd < q_order "
            "and 0 <= mkd < q_order }",
        ],
        """
            for tbox
                <> target_box_id  = target_boxes[tbox]

                <> tpt_begin = balls_near_box_starts[target_box_id]
                <> tpt_end   = balls_near_box_starts[target_box_id+1]

                <> box_level     = box_levels[target_box_id]
                <> box_mode_beg  = box_target_starts[target_box_id]
                <> n_box_modes   = box_target_counts_cumul[target_box_id]

                <> box_extent   = root_extent * (1.0 / (2**box_level))

                for iaxis
                    <> box_center[iaxis] = box_centers[iaxis, target_box_id] {dup=iaxis}
                end

                for tpt
                    <> target_point_id = balls_near_box_lists[tpt]

                    for iaxis
                        <> real_coord[iaxis] = TARGET_COORDS_ASSIGNMENT {dup=iaxis}
                    end

                    # Count how many times the potential is computed
                    multiplicity[target_point_id] = multiplicity[target_point_id] + 1

                    # Map target point to template box
                    for iaxis
                        <> tplt_coord[iaxis] = (real_coord[iaxis] - box_center[iaxis]
                            ) / box_extent + 0.5 {dup=iaxis}
                    end

                    # Precompute denominators
                    for iaxis
                        <> denom[iaxis] = 0.0 {id=reinit_denom,dup=iaxis}
                    end

                    for iaxis, mjd
                         <> diff[iaxis, mjd] = if( \
                                          tplt_coord[iaxis] == barycentric_lagrange_points[mjd], \
                                          1, \
                                          tplt_coord[iaxis] - barycentric_lagrange_points[mjd]) \
                                          {id=diff, dep=reinit_denom, dup=iaxis:mjd}
                         denom[iaxis] = denom[iaxis] + \
                                 barycentric_lagrange_weights[mjd] / diff[iaxis, mjd] \
                                 {id=denom, dep=diff, dup=iaxis:mjd}
                    end

                    for mid
                        # Find the coeff of each mode
                        <> mode_id      = box_mode_beg + mid
                        <> mode_id_user = user_mode_ids[mode_id]
                        <> mode_coeff   = potential[mode_id_user]

                        # Mode id in each direction
                        for iaxis
                            idx[iaxis] = MODE_INDEX_ASSIGNMENT {id=mode_indices,dup=iaxis}
                        end

                        # Interpolate mode value in each direction
                        for iaxis
                            <> numerator[iaxis] = (barycentric_lagrange_weights[idx[iaxis]]
                                                / diff[iaxis, idx[iaxis]]) {id=numer,dep=diff:mode_indices,dup=iaxis}
                            <> mode_val[iaxis] = numerator[iaxis] / denom[iaxis] {id=mode_val,dep=numer:denom,dup=iaxis}
                        end

                        # Fix when target point coincide with a quad point
                        for mkd, iaxis
                            mode_val[iaxis] = if(
                                    tplt_coord[iaxis] == barycentric_lagrange_points[mkd],
                                    if(mkd == idx[iaxis], 1, 0),
                                    mode_val[iaxis]) {id=fix_mode_val, dep=mode_val:mode_indices, dup=iaxis}
                        end

                        <> prod_mode_val = product(iaxis,
                            mode_val[iaxis]) {id=pmod,dep=fix_mode_val,dup=iaxis}

                    end

                    p_out[target_point_id] = p_out[target_point_id] + sum(mid,
                        mode_coeff * prod_mode_val
                        ) {id=p_out,dep=pmod}

                end

            end

            """.replace(
            "TARGET_COORDS_ASSIGNMENT", code_target_coords_assignment
        )
        .replace("MODE_INDEX_ASSIGNMENT", code_mode_index_assignment)
        .replace("Q_ORDER", "q_order"),
        [
            loopy.TemporaryVariable("idx", np.int32, "dim,"),
            # loopy.TemporaryVariable("denom", dtype, "dim,"),
            # loopy.TemporaryVariable("diff", dtype, "dim, q_order"),
            loopy.GlobalArg("box_centers", None, "dim, aligned_nboxes"),
            loopy.GlobalArg("balls_near_box_lists", None, None),
            loopy.ValueArg("aligned_nboxes", np.int32),
            loopy.ValueArg("dim", np.int32),
            loopy.ValueArg("q_order", np.int32),
            "...",
        ],
    )
    # }}} End loopy kernel for interpolation

    # loopy does not directly support object arrays
    if dim == 1:
        target_coords_knl_kwargs = {"target_coords_x": target_points[0]}
    elif dim == 2:
        target_coords_knl_kwargs = {
            "target_coords_x": target_points[0],
            "target_coords_y": target_points[1],
        }
    elif dim == 3:
        target_coords_knl_kwargs = {
            "target_coords_x": target_points[0],
            "target_coords_y": target_points[1],
            "target_coords_z": target_points[2],
        }
    else:
        raise NotImplementedError

    lpknl = loopy.set_options(lpknl, return_dict=True)
    lpknl = loopy.fix_parameters(lpknl, dim=int(dim), q_order=int(q_order))
    lpknl = loopy.split_iname(lpknl, "tbox", 128, outer_tag="g.0", inner_tag="l.0")
    evt, res_dict = lpknl(
        queue,
        p_out=pout,
        multiplicity=multiplicity,
        box_centers=tree.box_centers,
        box_levels=tree.box_levels,
        balls_near_box_starts=lbl_lookup.balls_near_box_starts,
        balls_near_box_lists=lbl_lookup.balls_near_box_lists,
        barycentric_lagrange_weights=blweights,
        barycentric_lagrange_points=blpoints,
        box_target_starts=tree.box_target_starts,
        box_target_counts_cumul=tree.box_target_counts_cumul,
        potential=potential,
        user_mode_ids=tree.user_source_ids,
        target_boxes=traversal.target_boxes,
        root_extent=tree.root_extent,
        n_tgt_boxes=len(traversal.target_boxes),
        **target_coords_knl_kwargs)

    assert pout is res_dict["p_out"]
    assert multiplicity is res_dict["multiplicity"]
    pout.add_event(evt)
    multiplicity.add_event(evt)

    return pout / multiplicity


# vim: filetype=pyopencl:fdm=marker



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import pyopencl as cl
import pytest
import volumential as vm
from volumential.table_manager import (
        NearFieldInteractionTableManager as NFTable)


def get_table(queue, q_order=1, dim=2):
    with NFTable("/tmp/volumential-tests.hdf5") as table_manager:
        table, _ = table_manager.get_table(
                dim, "Laplace",
                q_order=q_order, force_recompute=False, queue=queue)
    return table


def test_case_id(ctx_factory):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    table = get_table(queue)
    case_same_box = len(table.interaction_case_vecs) // 2
    assert list(table.interaction_case_vecs[case_same_box]) == [0, 0]


def test_get_table(ctx_factory):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    table = get_table(queue)
    assert table.dim == 2


def laplace_const_source_same_box(queue, q_order, dim=2):
    nft = get_table(queue, q_order, dim)

    n_pairs = nft.n_pairs
    n_q_points = nft.n_q_points
    pot = np.zeros(n_q_points)

    case_same_box = len(nft.interaction_case_vecs) // 2

    for source_mode_index in range(n_q_points):
        for target_point_index in range(n_q_points):
            pair_id = source_mode_index * n_q_points + target_point_index
            entry_id = case_same_box * n_pairs + pair_id
            # print(source_mode_index, target_point_index, pair_id, entry_id,
            # nft.data[entry_id])
            pot[target_point_index] += 1.0 * nft.data[entry_id]

    return pot


def laplace_cons_source_neighbor_box(queue, q_order, case_id, dim=2):
    nft = get_table(queue, q_order, dim)

    n_pairs = nft.n_pairs
    n_q_points = nft.n_q_points
    pot = np.zeros(n_q_points)

    for source_mode_index in range(n_q_points):
        for target_point_index in range(n_q_points):
            pair_id = source_mode_index * n_q_points + target_point_index
            entry_id = case_id * n_pairs + pair_id
            # print(source_mode_index, target_point_index, pair_id, entry_id,
            # nft.data[entry_id])
            pot[target_point_index] += 1.0 * nft.data[entry_id]

    return pot


def test_lcssb_1(ctx_factory):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    u = laplace_const_source_same_box(queue, 1)
    assert len(u) == 1


def interp_func(queue, q_order, coef, dim=2):
    nft = get_table(queue, q_order, dim)
    assert dim == 2

    modes = [nft.get_mode(i) for i in range(nft.n_q_points)]

    def func(x, y):
        z = np.zeros(np.array(x).shape)
        for i in range(nft.n_q_points):
            mode = modes[i]
            z += (coef[i] * mode(x, y)).reshape(z.shape)
        return z

    return func


def test_interp_func(longrun, ctx_factory):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    q_order = 3
    coef = np.ones(q_order ** 2)

    h = 0.1
    xx = yy = np.arange(-1.0, 1.0, h)
    xi, yi = np.meshgrid(xx, yy)
    func = interp_func(queue, q_order, coef)

    zi = func(xi, yi)

    assert np.allclose(zi, 1)


def direct_quad(source_func, target_point, dim=2):
    knl_func = vm.nearfield_potential_table.get_laplace(dim)

    def integrand(x, y):
        return source_func(x, y) * knl_func(x - target_point[0], y - target_point[1])
    import volumential.singular_integral_2d as squad
    integral, _ = squad.box_quad(
        func=integrand, a=0, b=1, c=0, d=1,
        singular_point=target_point, maxiter=1000)
    return integral


def drive_test_direct_quad_same_box(queue, q_order, dim=2):
    u = laplace_const_source_same_box(queue, q_order)
    func = interp_func(queue, q_order, u)
    nft = get_table(queue, q_order, dim)

    def const_one_source_func(x, y):
        return 1

    # print(nft.compute_table_entry(1341))
    # print(nft.compute_table_entry(nft.lookup_by_symmetry(1341)))

    for it in range(nft.n_q_points):
        target = nft.q_points[it]
        v1 = func(target[0], target[1])
        v2 = direct_quad(const_one_source_func, target)
        v3 = 0
        for ids in range(nft.n_q_points):
            mode = nft.get_mode(ids)
            vv = direct_quad(mode, target)
            print(ids, it, vv)
            v3 += vv

        print(target, v1, v2, v3)
        assert np.abs(v1 - v2) < 2e-6
        assert np.abs(v1 - v3) < 1e-6


@pytest.mark.parametrize("q_order", [1, ])
def test_direct_quad(q_order, ctx_factory):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    drive_test_direct_quad_same_box(queue, q_order)


@pytest.mark.parametrize("q_order", [2, 3, 4, 5])
def test_direct_quad_longrun(longrun, ctx_factory, q_order):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    drive_test_direct_quad_same_box(queue, q_order)


def test_case_ids(ctx_factory):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    table = get_table(queue)
    for i in range(len(table.interaction_case_vecs)):
        code = table.case_encode(
                table.interaction_case_vecs[i])
        assert table.case_indices[code] == i


def get_target_point(case_id, target_id, table):
    case_vec = table.interaction_case_vecs[case_id]
    center = np.array([0.5, 0.5]) + np.array(case_vec) * 0.25
    dist = np.max(np.abs(case_vec)) - 2
    if dist == 1:
        scale = 0.5
    elif dist == 2:
        scale = 1
    elif dist == 4:
        scale = 2
    dx = table.q_points[target_id][0] - 0.5
    dy = table.q_points[target_id][1] - 0.5
    target_point = np.array([center[0] + dx * scale, center[1] + dy * scale])
    return target_point


def test_get_neighbor_target_point(ctx_factory):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    table = get_table(queue)
    case_same_box = len(table.interaction_case_vecs) // 2
    for cid in range(len(table.interaction_case_vecs)):
        if cid == case_same_box:
            continue
        for tpid in range(table.n_q_points):
            pt = table.find_target_point(tpid, cid)
            pt2 = get_target_point(cid, tpid, table)
        assert np.allclose(pt, pt2)


def laplace_const_source_neighbor_box(queue, q_order, case_id, dim=2):
    nft = get_table(queue, q_order, dim)

    n_pairs = nft.n_pairs
    n_q_points = nft.n_q_points
    pot = np.zeros(n_q_points)

    for source_mode_index in range(n_q_points):
        for target_point_index in range(n_q_points):
            pair_id = source_mode_index * n_q_points + target_point_index
            entry_id = case_id * n_pairs + pair_id
            pot[target_point_index] += 1.0 * nft.data[entry_id]
    return pot


def drive_test_direct_quad_neighbor_box(queue, q_order, case_id, dim=2):
    u = laplace_const_source_neighbor_box(queue, q_order, case_id)
    nft = get_table(queue, q_order, dim)

    def const_one_source_func(x, y):
        return 1

    for it in range(nft.n_q_points):
        target = nft.find_target_point(it, case_id)
        v1 = u[it]
        v2 = direct_quad(const_one_source_func, target)
        v3 = 0
        for ids in range(nft.n_q_points):
            mode = nft.get_mode(ids)
            vv = direct_quad(mode, target)
            print(ids, it, vv)
            v3 += vv

        print(target, v1, v2, v3)
        assert np.abs(v1 - v2) < 2e-6
        assert np.abs(v1 - v3) < 1e-6


@pytest.mark.parametrize("q_order", [1, ])
def test_direct_quad_neighbor_box(ctx_factory, q_order):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    table = get_table(queue)
    for case_id in range(len(table.interaction_case_vecs)):
        drive_test_direct_quad_neighbor_box(queue, q_order, case_id)


@pytest.mark.parametrize("q_order", [2, ])
def test_direct_quad_neighbor_box_longrun(longrun, ctx_factory, q_order):
    cl_ctx = ctx_factory()
    queue = cl.CommandQueue(cl_ctx)
    table = get_table(queue)
    for case_id in range(len(table.interaction_case_vecs)):
        drive_test_direct_quad_neighbor_box(queue, q_order, case_id)


# fdm=marker:ft=pyopencl



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__doc__ = """
.. autoclass:: NearFieldEvalBase
   :members:
.. autoclass:: NearFieldFromCSR
   :members:
"""

import numpy as np
import loopy
import pyopencl as cl

from volumential.tools import KernelCacheWrapper

import logging

logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.INFO)

# {{{ near field eval base class


class NearFieldEvalBase(KernelCacheWrapper):
    """Base class of near-field evalulator.
    """

    default_name = "near_field_eval_base"

    def __init__(
        self,
        integral_kernel,
        table_data_shapes,
        potential_kind=1,
        options=[],
        name=None,
        device=None,
        **kwargs
    ):
        """potential_kind:

        1 - The (weakly singular) volume potentials.
        2 - The (hypersingular) inverse potentials, like the fractional Laplacian.
            Here, the fractional Laplacian is the inverse of the (weakly singular)
            Riesz potential operator.

        The two kinds share the same far-field code, but the second kind requires
        exterior_mode_nmlz when computing the list1 interactions.
        """

        self.integral_kernel = integral_kernel

        self.n_tables = table_data_shapes["n_tables"]
        self.n_q_points = table_data_shapes["n_q_points"]
        self.n_table_entries = table_data_shapes["n_table_entries"]
        self.potential_kind = potential_kind

        assert np.isreal(self.n_tables)
        assert np.isreal(self.n_q_points)
        assert np.isreal(self.n_table_entries)

        self.options = options
        self.name = name or self.default_name
        self.divice = device
        self.extra_kwargs = kwargs

        # Allow user to pass more tables to force using multiple tables
        # instead of performing kernel scaling
        if "infer_kernel_scaling" not in self.extra_kwargs:
            self.extra_kwargs["infer_kernel_scaling"] = self.n_tables == 1

        # Do not infer scaling rules when user defined rules are present
        if ("kernel_scaling_code" in self.extra_kwargs) or (
                "kernel_displacement_code" in self.extra_kwargs):
            self.extra_kwargs["infer_kernel_scaling"] = False
            # the two codes must be simultaneously given
            assert ("kernel_scaling_code" in self.extra_kwargs) and (
                "kernel_displacement_code" in self.extra_kwargs)

        self.kname = self.integral_kernel.__repr__()
        self.dim = self.integral_kernel.dim

    def get_cache_key(self):
        return (
            type(self).__name__,
            self.name,
            self.kname,
            "infer_scaling=" + str(self.extra_kwargs["infer_kernel_scaling"]),
        )


# }}} End near field eval base class

# {{{ eval from CSR data


class NearFieldFromCSR(NearFieldEvalBase):
    """Evaluate the near-field potentials from CSR representation of the tree.
    The class supports auto-scaling of simple kernels.
    """

    default_name = "near_field_from_csr"

    def codegen_vec_component(self, d=None):
        if d is None:
            dimension = self.dim - 1
        else:
            dimension = d
        return (
            "("
            + "(box_centers["
            + str(dimension)
            + ", target_box_id]"
            + "- box_centers["
            + str(dimension)
            + ", source_box_id]"
            + ") / sbox_extent * 4.0 + encoding_shift"
            + ")"
        )

    def codegen_vec_id(self):
        dim = self.dim
        code = "0.0"
        for d in range(dim):
            code = "(" + code + ") * encoding_base"
            code = code + "+" + self.codegen_vec_component(d)
        return code

    def codegen_compute_scaling(self, box_name='sbox'):
        """box_name: the name of the box whose extent is used.
        """
        if ("infer_kernel_scaling" in self.extra_kwargs) and (
            self.extra_kwargs["infer_kernel_scaling"]
        ):
            # Laplace 2D
            if self.kname == "LapKnl2D":
                logger.info("scaling for LapKnl2D")
                code = "BOX_extent * BOX_extent / \
                        (table_root_extent * table_root_extent)"

            elif self.kname in (
                    "AxisTargetDerivative(0, LapKnl2D)",
                    "AxisTargetDerivative(1, LapKnl2D)"):
                logger.info("scaling for Grad(LapKnl2D)")
                code = "BOX_extent / table_root_extent"

            # Constant 2D
            elif self.kname == "CstKnl2D":
                logger.info("scaling for CstKnl2D")
                code = "BOX_extent * BOX_extent / \
                        (table_root_extent * table_root_extent)"

            # Laplace 3D
            elif self.kname == "LapKnl3D":
                logger.info("scaling for Lapknl3D")
                code = "BOX_extent * BOX_extent / \
                        (table_root_extent * table_root_extent)"

            elif self.kname in (
                    "AxisTargetDerivative(0, LapKnl3D)",
                    "AxisTargetDerivative(1, LapKnl3D)",
                    "AxisTargetDerivative(2, LapKnl3D)"):
                logger.info("scaling for Grad(LapKnl3D)")
                code = "BOX_extent / table_root_extent"

            # Constant 3D
            elif self.kname == "CstKnl3D":
                logger.info("scaling for CstKnl3D")
                code = "BOX_extent * BOX_extent * BOX_extent / \
                        (table_root_extent * table_root_extent * table_root_extent)"

            else:
                logger.warn(
                    "Kernel not scalable and not using multiple tables, "
                    "to get correct results, please make sure that your "
                    "tree is uniform and only needs one table."
                )
                code = "1.0"

            return code.replace("BOX", box_name)

        elif "kernel_scaling_code" in self.extra_kwargs:
            # user-defined scaling rule
            assert isinstance(self.extra_kwargs['kernel_scaling_code'], str)
            logger.info("Using scaling rule %s for %s.",
                        self.extra_kwargs['kernel_scaling_code'], self.kname
                        )
            return self.extra_kwargs['kernel_scaling_code']
        else:
            logger.info("not scaling for " + self.kname)
            logger.info("(using multiple tables)")
            return "1.0"

    def codegen_compute_displacement(self, box_name='sbox'):
        if ("infer_kernel_scaling" in self.extra_kwargs) and (
            self.extra_kwargs["infer_kernel_scaling"]
        ):
            # Laplace 2D
            if self.kname == "LapKnl2D":
                logger.info("displacement for laplace 2D")
                s = "-0.5 / PI * scaling * \
                        log(BOX_extent / table_root_extent) * \
                        mode_nmlz[table_lev, sid]"
                import math
                code = s.replace("PI", str(math.pi))

            # Constant 2D
            elif self.kname == "CstKnl2D":
                logger.info("no displacement for CstKnl2D")
                code = "0.0"
            # Laplace 3D
            elif self.kname == "LapKnl3D":
                logger.info("no displacement for LapKnl3D")
                code = "0.0"
            # Constant 3D
            elif self.kname == "CstKnl3D":
                logger.info("no displacement for CstKnl3D")
                code = "0.0"
            else:
                logger.warn(
                    "Kernel not scalable and not using multiple tables, "
                    "to get correct results, please make sure that either "
                    "no displacement is needed, or the box "
                    "tree is uniform and only needs one table."
                )
                code = "0.0"
        elif "kernel_displacement_code" in self.extra_kwargs:
            # user-defined displacement rule
            assert isinstance(
                    self.extra_kwargs['kernel_displacement_code'], str)
            logger.info("Using displacement %s for %s.",
                        self.extra_kwargs['kernel_displacement_code'],
                        self.kname
                        )
            code = self.extra_kwargs['kernel_displacement_code']
        else:
            logger.info("no displacement for " + self.kname)
            logger.info("(using multiple tables)")
            code = "0.0"

        return code.replace("BOX", box_name)

    def codegen_get_table_level(self, box_name='sbox'):
        if ("infer_kernel_scaling" in self.extra_kwargs) and (
            self.extra_kwargs["infer_kernel_scaling"]
        ):
            if (
                self.kname == "LapKnl2D"
                or self.kname == "LapKnl3D"
                or self.kname == "CstKnl2D"
                or self.kname == "CstKnl3D"
            ):
                logger.info("scaling from table[0] for " + self.kname)
                code = "0.0"
            else:
                logger.warn(
                    "Kernel not scalable and not using multiple tables, "
                    "to get correct results, please make sure that your "
                    "tree is uniform and only needs one table."
                )
                code = "0.0"
        elif "kernel_scaling_code" in self.extra_kwargs:
            # Using custom scaling
            code = "0.0"
        else:
            logger.info("computing table level from box size")
            logger.info("(using multiple tables)")
            code = "log(table_root_extent / BOX_extent) / log(2.0)"

        return code.replace("BOX", box_name)

    def codegen_exterior_part(self):
        """Computes the exterior contribution. This is nonzero for
        inverse-type potentials like the fractional Laplacian.
        """
        if self.potential_kind == 1:
            return "0.0"
        elif self.potential_kind == 2:
            return "source_coefs[target_id] * ext_nmlz"
        else:
            raise ValueError("Unsupported potential kind %d" % self.potential_kind)

    def get_kernel(self):

        lpknl = loopy.make_kernel(  # NOQA
            [
                "{ [ tbox ] : 0 <= tbox < n_tgt_boxes }",
                "{ [ tid, sbox ] : 0 <= tid < n_box_targets and \
                        sbox_begin <= sbox < sbox_end }",
                "{ [ sid ] : 0 <= sid < n_box_sources }",
            ],
            """
            for tbox
                <> target_box_id    = target_boxes[tbox]
                <> box_target_beg   = box_target_starts[target_box_id]
                <> n_box_targets    = box_target_counts_cumul[target_box_id]

                <> sbox_begin = neighbor_source_boxes_starts[tbox]
                <> sbox_end   = neighbor_source_boxes_starts[tbox+1]

                <> tbox_level  = box_levels[target_box_id]
                <> tbox_extent = root_extent * (1.0 / (2**tbox_level))

                for tid
                    <> target_id = box_target_beg + tid
                end

                for tid, sbox
                    <> source_box_id  = source_boxes[sbox]
                    <> n_box_sources  = box_source_counts_cuml[source_box_id]
                    <> box_source_beg = box_source_starts[source_box_id]

                    <> sbox_level  = box_levels[source_box_id]
                    <> sbox_extent = root_extent * (1.0 / (2**sbox_level))

                    table_lev_tmp = GET_TABLE_LEVEL {id=tab_lev_tmp}
                    table_lev = round(table_lev_tmp) {id=tab_lev,dep=tab_lev_tmp}

                    vec_id_tmp = COMPUTE_VEC_ID {id=vec_id_tmp}
                    vec_id = round(vec_id_tmp) {id=vec_id,dep=vec_id_tmp}
                    <> case_id = case_indices[vec_id] {dep=vec_id}

                    <> scaling = COMPUTE_SCALING

                    for sid

                        <> tgt_scaling = COMPUTE_TGT_SCALING
                        <> tgt_displacement = COMPUTE_TGT_DISPLACEMENT
                        tgt_table_lev_tmp = GET_TGT_TABLE_LEVEL {id=tgttab_lev_tmp}
                        tgt_table_lev = round(tgt_table_lev_tmp) \
                                {id=tgttab_lev,dep=tgttab_lev_tmp}
                        <> ext_nmlz = exterior_mode_nmlz[tgt_table_lev, tid] \
                                * tgt_scaling + tgt_displacement \
                                {id=extnmlz,dep=tgttab_lev}

                        <> source_id = box_source_beg + sid
                        <> pair_id = sid * n_box_targets + tid
                        <> entry_id = case_id * \
                                      (n_box_targets * n_box_sources) \
                                      + pair_id

                        <> displacement = COMPUTE_DISPLACEMENT

                        <> integ = table_data[table_lev, entry_id] * scaling \
                                   + displacement {id=integ,dep=tab_lev}
                        # <> source_id_tree = user_source_ids[source_id]
                        <> coef = source_coefs[source_id] {id=coef}

                        # <> target_id_user = sorted_target_ids[target_id]

                        #db_table_lev[target_id] = table_lev_tmp {dep=tab_lev}
                        #db_case_id[target_id] = case_id
                        #db_vec_id[target_id] = vec_id
                        #db_n_box_targets[target_id] = n_box_targets
                        #db_n_box_sources[target_id] = n_box_sources
                        #db_entry_id[target_id] = entry_id

                    end
                end

                for tid

                    result[target_id] = sum((sbox, sid),
                        coef * integ) + EXTERIOR_PART {dep=integ:coef:extnmlz}

                    # Try inspecting case_id if something goes wrong
                    # (like segmentation fault) and look for -1's
                    # result[target_id] = min((sbox, sid), case_id)
                    # result[target_id] = vec_id_tmp

                end
            end
            """
            .replace("COMPUTE_VEC_ID", self.codegen_vec_id())
            .replace("COMPUTE_SCALING", self.codegen_compute_scaling())
            .replace("COMPUTE_DISPLACEMENT", self.codegen_compute_displacement())
            .replace("COMPUTE_TGT_SCALING", self.codegen_compute_scaling('tbox'))
            .replace("COMPUTE_TGT_DISPLACEMENT",
                    self.codegen_compute_displacement('tbox'))
            .replace("GET_TABLE_LEVEL", self.codegen_get_table_level())
            .replace("GET_TGT_TABLE_LEVEL", self.codegen_get_table_level('tbox'))
            .replace("EXTERIOR_PART", self.codegen_exterior_part()),
            [
                loopy.TemporaryVariable("vec_id", np.int32),
                loopy.TemporaryVariable("vec_id_tmp", np.float64),
                loopy.TemporaryVariable("table_lev", np.int32),
                loopy.TemporaryVariable("table_lev_tmp", np.float64),
                loopy.TemporaryVariable("tgt_table_lev", np.int32),
                loopy.TemporaryVariable("tgt_table_lev_tmp", np.float64),
                loopy.ValueArg("encoding_base", np.int32),
                loopy.GlobalArg("mode_nmlz", np.float64, "n_tables, n_q_points"),
                loopy.GlobalArg("exterior_mode_nmlz", np.float64,
                    "n_tables, n_q_points"),
                loopy.GlobalArg("table_data", np.float64,
                    "n_tables, n_table_entries"),
                loopy.GlobalArg("source_boxes", np.int32, "n_source_boxes"),
                loopy.GlobalArg("box_centers", None, "dim, aligned_nboxes"),
                loopy.ValueArg("aligned_nboxes", np.int32),
                loopy.ValueArg("table_root_extent", np.float64),
                loopy.ValueArg(
                    "dim, n_source_boxes, n_tables, " "n_q_points, n_table_entries",
                    np.int32,
                ),
                "...",
            ],
            name="near_field",
            lang_version=(2018, 2)
        )

        # print(lpknl)
        # lpknl = loopy.set_options(lpknl, write_code=True)

        lpknl = loopy.set_options(lpknl, return_dict=True)

        return lpknl

    def get_cache_key(self):
        return (
            type(self).__name__,
            self.name,
            self.kname,
            "potential_kind=%d" % self.potential_kind,
            "infer_scaling=" + str(self.extra_kwargs["infer_kernel_scaling"]),
            "scaling_policy=" + self.codegen_compute_scaling(),
            "displacement_policy=" + self.codegen_compute_displacement(),
        )

    def get_optimized_kernel(self, ncpus=None):
        if ncpus is None:
            import multiprocessing
            # NOTE: this detects the number of logical cores, disable hyperthreading
            # for the optimal performance.
            ncpus = multiprocessing.cpu_count()
        knl = self.get_kernel()
        knl = loopy.split_iname(knl, "tbox", ncpus, inner_tag="g.0")
        return knl

    def __call__(self, queue, **kwargs):
        knl = self.get_cached_optimized_kernel()

        result = kwargs.pop("result")
        box_centers = kwargs.pop("box_centers")
        box_levels = kwargs.pop("box_levels")
        box_source_counts_cumul = kwargs.pop("box_source_counts_cumul")
        box_source_starts = kwargs.pop("box_source_starts")
        box_target_counts_cumul = kwargs.pop("box_target_counts_cumul")
        box_target_starts = kwargs.pop("box_target_starts")
        case_indices = kwargs.pop("case_indices")
        encoding_base = kwargs.pop("encoding_base")
        encoding_shift = kwargs.pop("encoding_shift")
        mode_nmlz_combined = kwargs.pop("mode_nmlz_combined")
        exterior_mode_nmlz_combined = kwargs.pop("exterior_mode_nmlz_combined")
        root_extent = kwargs.pop("root_extent")
        table_root_extent = kwargs.pop("table_root_extent")
        neighbor_source_boxes_starts = kwargs.pop("neighbor_source_boxes_starts")
        neighbor_source_boxes_lists = kwargs.pop("neighbor_source_boxes_lists")
        mode_coefs = kwargs.pop("mode_coefs")
        table_data_combined = kwargs.pop("table_data_combined")
        target_boxes = kwargs.pop("target_boxes")

        integral_kernel_init_kargs = {
                name: val
                for name, val in zip(
                    self.integral_kernel.init_arg_names,
                    self.integral_kernel.__getinitargs__())
                }

        # help loopy's type inference
        for key, val in integral_kernel_init_kargs.items():
            if isinstance(val, int):
                integral_kernel_init_kargs[key] = np.int32(val)
            if isinstance(val, float):
                integral_kernel_init_kargs[key] = np.float64(val)

        extra_knl_args_from_init = {}
        for key, val in integral_kernel_init_kargs.items():
            if key in knl.arg_dict:
                extra_knl_args_from_init[key] = val

        evt, res = knl(
            queue,
            result=result,
            # db_table_lev=np.zeros(out_pot.shape),
            # db_case_id=np.zeros(out_pot.shape),
            # db_vec_id=np.zeros(out_pot.shape),
            # db_n_box_sources=np.zeros(out_pot.shape),
            # db_n_box_targets=np.zeros(out_pot.shape),
            # db_entry_id=np.zeros(out_pot.shape),
            box_centers=box_centers,
            box_levels=box_levels,
            box_source_counts_cuml=box_source_counts_cumul,
            box_source_starts=box_source_starts,
            box_target_counts_cumul=box_target_counts_cumul,
            box_target_starts=box_target_starts,
            case_indices=case_indices,
            n_tables=self.n_tables,
            n_table_entries=self.n_table_entries,
            n_q_points=self.n_q_points,
            encoding_base=encoding_base,
            encoding_shift=encoding_shift,
            mode_nmlz=mode_nmlz_combined,
            exterior_mode_nmlz=exterior_mode_nmlz_combined,
            n_tgt_boxes=len(target_boxes),
            neighbor_source_boxes_starts=neighbor_source_boxes_starts,
            root_extent=root_extent,
            source_boxes=neighbor_source_boxes_lists,
            n_source_boxes=len(neighbor_source_boxes_lists),
            source_coefs=mode_coefs,
            table_data=table_data_combined,
            target_boxes=target_boxes,
            table_root_extent=table_root_extent,
            **extra_knl_args_from_init
            )

        res['result'].add_event(evt)
        if isinstance(result, cl.array.Array):
            assert result is res["result"]
        else:
            assert isinstance(result, np.ndarray)
            result = res['result'].get(queue)

        queue.finish()
        logger.info("list1 evaluation finished")

        # check for data integrity
        # if not (np.max(np.abs(out_pot.get()))) < 100:
        #    import pudb; pu.db
        return result, evt


# }}} End eval from CSR data



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import volumential.singular_integral_2d as sint


def test_quadrature_1d_interval():
    integrand = np.sin
    val, err = sint.quad(
        func=integrand,
        a=-np.pi,
        b=np.pi,
        args=(),
        tol=1e-8,
        rtol=1e-8,
        maxiter=50,
        vec_func=False,
        miniter=1,
    )
    assert np.abs(val) < 1e-8
    assert err < 1e-8


def test_quadrature_2d_box():
    def integrand(x, y):
        return np.sin(10 * x) * np.cos(3 * y)

    val, err = sint.qquad(
        func=integrand,
        a=-np.pi,
        b=np.pi,
        c=-np.pi,
        d=np.pi,
        args=(),
        tol=1e-8,
        rtol=1e-8,
        maxitero=50,
        maxiteri=50,
        vec_func=True,
        minitero=1,
        miniteri=1,
    )
    assert np.abs(val) < 1e-8
    assert err < 1e-8


def test_affine_mapping():
    tria1 = ((1, 1), (2, 3), (0, 4))
    tria2 = ((0, 0), (1, 0), (0, 1))
    mp, J, imp, iJ = sint.solve_affine_map_2d(tria1, tria2)
    assert np.isclose(J * iJ, 1)
    for i in range(3):
        assert np.allclose(mp(tria1[i]), tria2[i])
        assert np.allclose(imp(tria2[i]), tria1[i])


def test_is_in_t():
    p1 = (0.1, 0.1)
    p2 = (1.1, 0.1)
    p3 = (-0.1, 0.1)
    p4 = (0.1, -0.7)
    assert sint.is_in_t(p1)
    assert not sint.is_in_t(p2)
    assert not sint.is_in_t(p3)
    assert not sint.is_in_t(p4)


def test_is_in_r():
    p1 = (0.1, 0.1)
    p2 = (1.1, 0.1)
    p3 = (-0.1, 0.1)
    p4 = (0.1, -0.7)
    assert sint.is_in_r(p1)
    assert not sint.is_in_r(p2)
    assert not sint.is_in_r(p3)
    assert not sint.is_in_r(p4)


def test_tria2rect_map():
    t2r, J_t2r, r2t, J_r2t = sint.tria2rect_map_2d()
    assert np.allclose(t2r((0, 0)), (0, 0))
    assert np.allclose(t2r((1, 0)), (1, 0))
    assert np.allclose(t2r((0, 1)), (1, np.pi / 2))
    assert np.allclose(r2t((0, 0)), (0, 0))
    assert np.allclose(r2t((1, 0)), (1, 0))
    assert np.allclose(r2t((1, np.pi / 2)), (0, 1))
    assert np.allclose(r2t((0, np.pi / 2)), (0, 0))


def test_tria2rect_jacobian():
    t2r, J_t2r, r2t, J_r2t = sint.tria2rect_map_2d()
    p1 = (0.1, 0.1)
    assert np.isclose(J_t2r(p1) * J_r2t(t2r(p1)), 1)
    assert np.isclose(J_r2t(p1) * J_t2r(r2t(p1)), 1)


def run_test_tria_quad(func, region, exact):
    val, err = sint.tria_quad(
        func,
        region,
        args=(),
        tol=1e-10,
        rtol=1e-10,
        maxiter=50,
        vec_func=True,
        miniter=1,
    )

    assert np.isclose(exact, val, atol=1e-8)
    assert np.isclose(err, 0, atol=1e-8)


def test_tria_quad_1():
    def const_func(x, y):
        return 1

    # area = 0.5 * (10 * 9) = 45
    region = ((2.33, 1), (12.33, 1), (5, 10))

    run_test_tria_quad(const_func, region, 45)


def test_tria_quad_2():
    def greens_func(x, y, x0, y0):
        return -1 / (2 * np.pi) * np.log((x - x0) ** 2 + (y - y0) ** 2) * 0.5

    region = ((2.33, 1), (12.33, 1), (5, 10))

    val, err = sint.tria_quad(
        greens_func,
        region,
        args=region[0],
        tol=1e-8,
        rtol=1e-8,
        maxiter=80,
        vec_func=True,
        miniter=1,
    )

    assert np.isclose(err, 0, atol=1e-6)
    assert np.isfinite(val)


def test_box_quad_1():
    def const_func(x, y):
        return 1

    a = 1
    b = 2
    c = 23
    d = 40
    area = (b - a) * (d - c)
    sp = ((a + b) / 2, (c + d) / 2)

    val, err = sint.box_quad(
        const_func,
        a,
        b,
        c,
        d,
        sp,
        args=(),
        tol=1e-8,
        rtol=1e-8,
        maxiter=50,
        vec_func=True,
        miniter=1,
    )

    assert np.isclose(val, area, atol=1e-8)
    assert np.isclose(err, 0, atol=1e-8)


def test_box_quad_2():
    def const_func(x, y):
        return 1

    a = 1
    b = 2
    c = 23
    d = 40
    area = (b - a) * (d - c)
    sp = (0, 0)

    val, err = sint.box_quad(
        const_func,
        a,
        b,
        c,
        d,
        sp,
        args=(),
        tol=1e-8,
        rtol=1e-8,
        maxiter=50,
        vec_func=True,
        miniter=1,
    )

    print(area, val)
    print(err)
    assert np.isclose(val, area, atol=1e-8)
    assert np.isclose(err, 0, atol=1e-8)


def test_box_quad_3(longrun):
    def greens_func(x, y, x0, y0):
        return -1 / (2 * np.pi) * np.log((x - x0) ** 2 + (y - y0) ** 2) * 0.5

    a = 0
    b = 1
    c = 0
    d = 1
    # by symmetry, all four singular integrals should be equal
    sps = [(0.1, 0.1), (0.9, 0.1), (0.9, 0.9), (0.1, 0.9)]

    val = np.zeros(len(sps))
    err = np.zeros(len(sps))

    for i in range(len(sps)):
        val[i], err[i] = sint.box_quad(
            greens_func,
            a,
            b,
            c,
            d,
            sps[i],
            args=sps[i],
            tol=1e-8,
            rtol=1e-8,
            maxiter=50,
            vec_func=True,
            miniter=1,
        )

    val = val - np.mean(val)

    assert np.allclose(val, 0, atol=1e-8)
    assert np.allclose(err, 0, atol=1e-8)


# vim: filetype=pyopencl.python:fdm=marker



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import numpy as np
import scipy.integrate as sint
import scipy.integrate.quadrature as quad  # noqa
import scipy as sp

__doc__ = """The 2D singular integrals are computed using the transform
described in http://link.springer.com/10.1007/BF00370482.

.. autofunction:: box_quad
"""

logger = logging.getLogger(__name__)

# {{{ quadrature on rectangular box with no singularity

quad_points_x = np.array([])
quad_points_y = np.array([])
quad_weights = np.array([])


def update_qquad_leggauss_formula(deg1, deg2):

    x1, w1 = np.polynomial.legendre.leggauss(deg1)
    x1 = (x1 + 1) / 2
    w1 = w1 / 2

    x2, w2 = np.polynomial.legendre.leggauss(deg2)
    x2 = (x2 + 1) / 2
    w2 = w2 / 2

    quad_points_x, quad_points_y = np.meshgrid(x1, x2)
    ww1, ww2 = np.meshgrid(w1, w2)
    global quad_weights
    quad_weights = ww1 * ww2


def qquad(
    func,
    a,
    b,
    c,
    d,
    args=(),
    tol=1.49e-08,
    rtol=1.49e-08,
    maxitero=50,
    maxiteri=50,
    vec_func=False,
    minitero=1,
    miniteri=1,
    method="Adaptive",
):
    """Computes a (tensor product) double integral.

    Integrate func on [a, b]X[c, d] using Gaussian quadrature with absolute
    tolerance tol.

    :param func: A double variable Python function or method to integrate.
    :type func: function.
    :param a: Lower-left corner of integration region.
    :type a: float.
    :param b: Lower-right corner of integration region.
    :type b: float.
    :param c: Upper-left corner of integration region.
    :type c: float.
    :param d: Upper-right corner of integration region.
    :type d: float.
    :param args: Extra arguments to pass to function.
    :type args: tuple, optional.
    :param tol: rtol Iteration stops when error between last two iterates is
                less than tol OR the relative change is less than rtol.
    :type tol: float, optional.
    :param rtol: Iteration stops when error between last two iterates is less
                than tol OR the relative change is less than rtol.
    :type rtol: float, optional.
    :param maxitero: Maximum order of outer Gaussian quadrature.
    :type maxitero: int, optional.
    :param maxiteri: Maximum order of inner Gaussian quadrature.
    :type maxiteri: int, optional.
    :param vec_func: True if func handles arrays as arguments (is a "vector"
                function). Default is True.
    :type vec_func: bool, optional.
    :param minitero: Minimum order of outer Gaussian quadrature.
    :type minitero: int, optional.
    :param miniteri: Minimum order of inner Gaussian quadrature.
    :type miniteri: int, optional.

    :returns:
        - **val**: Gaussian quadrature approximation (within tolerance) to integral.
        - **err**: Difference between last two estimates of the integral.
    :rtype: tuple(float,float).
    """

    l1 = b - a
    l2 = d - c
    assert l1 > 0
    assert l2 > 0
    toli = tol / l2
    rtoli = rtol / l2

    if method == "Adaptive":

        # Using lambda for readability
        def outer_integrand(y):
            return sint.quadrature(  # NOQA
                lambda x: func(x, y, *args),
                a,
                b,
                (),
                toli,
                rtoli,
                maxiteri,
                vec_func,
                miniteri,
            )[
                0
            ]  # NOQA

        # Is there a simple way to retrieve err info from the inner quad calls?

        val, err = sint.quadrature(
            outer_integrand, c, d, (), tol, rtol, maxitero, vec_func, minitero
        )

    elif method == "Gauss":

        # Gauss quadrature with orders equal to maxiters

        # Using lambda for readability
        def outer_integrand(y):
            return sp.integrate.fixed_quad(  # NOQA
                np.vectorize(lambda x: func(x, y, *args)), a, b, (), maxiteri
            )[
                0
            ]  # NOQA

        # Is there a simple way to retrieve err info from the inner quad calls?

        val, err = sp.integrate.fixed_quad(
            np.vectorize(outer_integrand), c, d, (), maxitero
        )

        assert err is None

    else:

        raise NotImplementedError("Unsupported quad method: " + method)

    return (val, err)


# }}}

# {{{ affine mappings


def solve_affine_map_2d(source_tria, target_tria):
    """Computes the affine map and its inverse that maps the source_tria to
       target_tria.

       :param source_tria: The triangle to be mapped.
       :type source_tria:
            tuple(tuple(float,float),tuple(float,float),tuple(float,float)).
       :param target_tria: The triangle to map to.
       :type target_tria:
            tuple(tuple(float,float),tuple(float,float),tuple(float,float)).

       :returns:
        - **mapping**: the forward map.
        - **J**: the Jacobian.
        - **invmap**: the inverse map.
        - **invJ**: the Jacobian of inverse map.
       :rtype:
        tuple(lambda, float, lambda, float)
    """
    assert len(source_tria) == 3
    for p in source_tria:
        assert len(p) == 2

    assert len(target_tria) == 3
    for p in target_tria:
        assert len(p) == 2

    # DOFs: A11, A12, A21, A22, b1, b2
    rhs = np.array(
        [
            target_tria[0][0],
            target_tria[0][1],
            target_tria[1][0],
            target_tria[1][1],
            target_tria[2][0],
            target_tria[2][1],
        ]
    )

    coef = np.array(
        [
            [source_tria[0][0], source_tria[0][1], 0, 0, 1, 0],
            [0, 0, source_tria[0][0], source_tria[0][1], 0, 1],
            [source_tria[1][0], source_tria[1][1], 0, 0, 1, 0],
            [0, 0, source_tria[1][0], source_tria[1][1], 0, 1],
            [source_tria[2][0], source_tria[2][1], 0, 0, 1, 0],
            [0, 0, source_tria[2][0], source_tria[2][1], 0, 1],
        ]
    )

    # x, residuals, _, _ = np.linalg.lstsq(coef, rhs)
    # assert (np.allclose(residuals, 0))
    try:
        x = np.linalg.solve(coef, rhs)
    except np.linalg.linalg.LinAlgError:
        print("")
        print("source:", source_tria)
        print("target:", target_tria)
        raise SystemExit("Error: Singular source triangle encountered")
    assert len(x) == 6
    assert np.allclose(np.dot(coef, x), rhs)

    a = np.array([[x[0], x[1]], [x[2], x[3]]])
    b = np.array([x[4], x[5]])

    # Using default value is the idiomatic way to "capture by value"
    mapping = lambda x, a=a, b=b: a.dot(np.array(x)) + b  # NOQA
    jacob = np.linalg.det(a)

    inva = np.linalg.inv(a)
    invb = -inva.dot(b)
    invmap = lambda x, a=inva, b=invb: inva.dot(np.array(x)) + invb  # NOQA
    inv_jacob = np.linalg.det(inva)

    assert np.abs(jacob * inv_jacob - 1) < 1e-12

    return (mapping, jacob, invmap, inv_jacob)


# }}}

# {{{ standard-triangle-to-rectangle mappings


def tria2rect_map_2d():
    """Returns the mapping and its inverse that maps a template triangle to a
       template rectangle.

            - Template triangle [T]: (0,0)--(1,0)--(0,1)--(0,0)
            - Template rectangle [R]: (0,0)--(1,0)--(1,pi/2)--(0,pi/2)--(0,0)

       :returns: The mapping, its Jacobian, its inverse, and the Jacobian of its
                inverse. Note that the Jacobians are returned as lambdas since
                they are not constants.
       :rtype: tuple(lambda, lambda, lambda, lambda)
    """

    # (x,y) --> (rho, theta): T --> R
    def mapping(x):
        return (x[0] + x[1], np.arctan2(np.sqrt(x[1]), np.sqrt(x[0])))

    def jacob(x):
        return 1 / (2 * np.sqrt(x[0] * x[1]))

    # x = rho * cos^2(theta), y = rho * sin^2(theta)
    # J = rho * sin(2*theta)
    def invmap(u):
        return (u[0] * (np.cos(u[1]) ** 2), u[0] * (np.sin(u[1]) ** 2))

    def inv_jacob(u):
        return u[0] * np.sin(2 * u[1])

    return (mapping, jacob, invmap, inv_jacob)


def is_in_t(pt):
    """Checks if a point is in the template triangle T.

       :param pt: The point to be checked.
       :type pt: tuple(float,float).

       :returns: True if pt is in T.
       :rtype: bool.
    """
    flag = True
    if pt[0] < 0 or pt[1] < 0:
        flag = False
    if pt[0] + pt[1] > 1:
        flag = False
    return flag


def is_in_r(pt, a=0, b=1, c=0, d=np.pi / 2):
    """Checks if a point is in the (template) rectangle R.

       :param pt: The point to be checked.
       :type pt: tuple(float,float).

       :returns: True if pt is in R.
       :rtype: bool.
    """
    flag = True
    if pt[0] < a or pt[1] < c:
        flag = False
    if pt[0] > b or pt[1] > d:
        flag = False
    return flag


# }}}

# {{{ quadrature on arbitrary triangle


def is_collinear(p0, p1, p2):
    # v1 = p0 --> p1
    x1, y1 = p1[0] - p0[0], p1[1] - p0[1]
    # v2 = p0 --> p2
    x2, y2 = p2[0] - p0[0], p2[1] - p0[1]
    # v1 cross v2 == 0 <==> collinearity
    return np.abs(x1 * y2 - x2 * y1) < 1e-16


def is_positive_triangle(tria):
    p0 = tria[0]
    p1 = tria[1]
    p2 = tria[2]
    # v1 = p0 --> p1
    x1, y1 = p1[0] - p0[0], p1[1] - p0[1]
    # v2 = p0 --> p2
    x2, y2 = p2[0] - p0[0], p2[1] - p0[1]
    # v1 cross v2 > 0 <==> is positive
    return (x1 * y2 - x2 * y1) > 0


def tria_quad(
    func,
    tria,
    args=(),
    tol=1.49e-08,
    rtol=1.49e-08,
    maxiter=50,
    vec_func=True,
    miniter=1,
):
    """Computes a double integral on a general triangular region.

    Integrate func on tria by transforming the region into a rectangle and
    using Gaussian quadrature with absolute tolerance tol.

    The integrand, func, is allowed to have singularity at most $O(r)$ at the
    first virtex of the tiangle. It is okay if func does not evaluate at the
    singular point. This function handles that automatically.

    :param func: A double variable Python function or method to integrate.
    :type func: function.
    :param tria: The triangular region to do quadrature.
    :type tria:
        tuple(tuple(float,float), tuple(float,float), tuple(float,float)).
    :param args: Extra arguments to pass to function.
    :type args: tuple, optional.
    :param tol: rtol Iteration stops when error between last two iterates is
        less than tol OR the relative change is less than rtol.
    :type tol: float, optional.
    :param rtol: Iteration stops when error between last two iterates is less
        than tol OR the relative change is less than rtol.
    :type rtol: float, optional.
    :param maxiter: Maximum order of Gaussian quadrature.
    :type maxiter: int, optional.
    :param vec_func: True if func handles arrays as arguments
        (is a "vector" function). Default is True.
    :type vec_func: bool, optional.
    :param miniter: Minimum order of Gaussian quadrature.
    :type miniter: int, optional.

    :returns:
        - **val**: Gaussian quadrature approximation (within tolerance)
            to integral.
        - **err**: Difference between last two estimates of the integral.
    :rtype: tuple(float,float).
    """

    assert len(tria) == 3
    for p in tria:
        assert len(p) == 2

    # Handle degenerate triangles
    if is_collinear(*tria):
        return (0.0, 0.0)

    # The function must be regular at the last two vertices
    assert np.isfinite(func(tria[1][0], tria[1][1], *args))
    assert np.isfinite(func(tria[2][0], tria[2][1], *args))

    # Solve for transforms
    template_tria = ((0, 0), (1, 0), (0, 1))
    afmp, j_afmp, inv_afmp, j_inv_afmp = solve_affine_map_2d(tria, template_tria)
    nlmp, j_nlmp, inv_nlmp, j_inv_nlmp = tria2rect_map_2d()

    # tria --> rect
    def mapping(x, y):
        return nlmp(afmp((x, y)))

    def jacobian(x, y):
        return j_afmp * j_nlmp(afmp((x, y)))

    # rect --> tria
    def inv_mapping(rho, theta):
        return inv_afmp(inv_nlmp((rho, theta)))

    def inv_jacobian(rho, theta):
        return j_inv_afmp * j_inv_nlmp((rho, theta))

    # Transformed function is defined on [0,1]X[0,pi/2]
    def transformed_func(rho, theta):
        preimage = inv_mapping(rho, theta)
        return func(preimage[0], preimage[1], *args)

    # Transformed function, when multiplied by jacobian, should have no
    # singularity (numerically special treatment still needed)

    # integrand = func * jacobian

    def integrand(rho, theta):
        prior = transformed_func(rho, theta) * inv_jacobian(rho, theta)
        # If something blows up, it is near the singular point
        if ~np.isfinite(prior):
            assert rho < 1e-3
            assert inv_jacobian(rho, theta) < 1e-6
            prior = 0
        return prior

    return qquad(
        func=integrand,
        a=0,
        b=1,
        c=0,
        d=np.pi / 2,
        args=(),
        tol=tol,
        rtol=rtol,
        maxiteri=maxiter,
        maxitero=maxiter,
        vec_func=False,
        miniteri=miniter,
        minitero=miniter,
    )


# }}}

# {{{ quadrature on a 2d box with a singular point inside


def box_quad(
    func,
    a,
    b,
    c,
    d,
    singular_point,
    args=(),
    tol=1.49e-08,
    rtol=1.49e-08,
    maxiter=50,
    vec_func=True,
    miniter=1,
):
    """Computes a (tensor product) double integral, with the integrand being
        singular at some point inside the region.

    Integrate func on [a, b]X[c, d] using transformed Gaussian quadrature with
    absolute tolerance tol.

    :param func: A double variable Python function or method to integrate.
    :type func: function.
    :param a: Lower-left corner of integration region.
    :type a: float.
    :param b: Lower-right corner of integration region.
    :type b: float.
    :param c: Upper-left corner of integration region.
    :type c: float.
    :param d: Upper-right corner of integration region.
    :type d: float.
    :param singular_point: The singular point of the integrand func.
    :type singular_point: tuple(float, float).
    :param args: Extra arguments to pass to function.
    :type args: tuple, optional.
    :param tol: rtol Iteration stops when error between last two iterates is
                less than tol OR the relative change is less than rtol.
    :type tol: float, optional.
    :param rtol: Iteration stops when error between last two iterates is less
                than tol OR the relative change is less than rtol.
    :type rtol: float, optional.
    :param maxiter: Maximum order of Gaussian quadrature.
    :type maxiter: int, optional.
    :param vec_func: True if func handles arrays as arguments (is a "vector"
                function). Default is True.
    :type vec_func: bool, optional.
    :param miniter: Minimum order of Gaussian quadrature.
    :type miniter: int, optional.

    :returns:
        - **val**: Gaussian quadrature approximation (within tolerance) to integral.
        - **err**: Difference between last two estimates of the integral.
    :rtype: tuple(float,float).
    """
    box = ((a, c), (b, c), (b, d), (a, d))

    if not isinstance(singular_point, tuple):
        singular_point = (singular_point[0], singular_point[1])

    # When singular point is outside, project it onto the box bounday
    # This can import speed by not integrating around the actual singularity
    # when not necessary. (The splitting is still needed since it can be quite
    # close to singular).
    singular_point = (max(singular_point[0], a), max(singular_point[1], c))
    singular_point = (min(singular_point[0], b), min(singular_point[1], d))

    return quadri_quad(
        func, box, singular_point, args, tol, rtol, maxiter, vec_func, miniter
    )


# quadrilateral
def quadri_quad(
    func,
    quadrilateral,
    singular_point,
    args=(),
    tol=1.49e-08,
    rtol=1.49e-08,
    maxiter=50,
    vec_func=True,
    miniter=1,
):
    """Computes a double integral over a (non-twisted) quadrilateral, with the
        integrand being singular at some point inside the region.

    Integrate func on [a, b]X[c, d] using transformed Gaussian quadrature with
    absolute tolerance tol.

    :param func: A double variable Python function or method to integrate.
    :type func: function.
    :param quadrilateral: The integration region.
    :type quadrilateral: tuple(tuple(float,float),
        tuple(float,float), tuple(float,float), tuple(float,float)).
    :param singular_point: The singular point of the integrand func.
    :type singular_point: tuple(float, float).
    :param args: Extra arguments to pass to function.
    :type args: tuple, optional.
    :param tol: rtol Iteration stops when error between last two iterates is
                less than tol OR the relative change is less than rtol.
    :type tol: float, optional.
    :param rtol: Iteration stops when error between last two iterates is less
                than tol OR the relative change is less than rtol.
    :type rtol: float, optional.
    :param maxiter: Maximum order of Gaussian quadrature.
    :type maxiter: int, optional.
    :param vec_func: True if func handles arrays as arguments (is a "vector"
                function). Default is True.
    :type vec_func: bool, optional.
    :param miniter: Minimum order of Gaussian quadrature.
    :type miniter: int, optional.

    :returns:
        - **val**: Gaussian quadrature approximation (within tolerance) to integral.
        - **err**: Difference between last two estimates of the integral.
    :rtype: tuple(float,float).
    """
    assert len(quadrilateral) == 4
    for p in quadrilateral:
        assert len(p) == 2

    # split the quadrilateral into four triangles
    trias = [
        (singular_point, quadrilateral[0], quadrilateral[1]),
        (singular_point, quadrilateral[1], quadrilateral[2]),
        (singular_point, quadrilateral[2], quadrilateral[3]),
        (singular_point, quadrilateral[3], quadrilateral[0]),
    ]

    for tria in trias:
        if not is_positive_triangle(tria):
            assert is_collinear(*tria)

    val = np.zeros(4)
    err = np.zeros(4)

    for i in range(4):
        val[i], err[i] = tria_quad(
            func, trias[i], args, tol, rtol, maxiter, vec_func, miniter
        )

    integral = np.sum(val)
    error = np.linalg.norm(err)

    return (integral, error)


# }}}
'''
class DesingularizationMapping:
    def __init__(self, nquad_points_1d):

def build_singular_box_quadrature(
        kernel,
        desing_mapping,

        ):
    """
    :arg kernel: an instance of :class:`sumpy.kernel.Kernel`
    :arg desing_mapping: an instance of :class:`sumpy.kernel.Kernel`
    """
'''

# vim: filetype=pyopencl.python:fdm=marker



""" This example evaluates the volume potential over
    [-1,1]^2 with the Laplace kernel.
"""
from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

logger = logging.getLogger(__name__)

if 0:
    # verbose
    logging.basicConfig(level=logging.INFO)
else:
    # clean
    logging.basicConfig(level=logging.CRITICAL)

import numpy as np
import pyopencl as cl
from volumential.tools import ScalarFieldExpressionEvaluation as Eval

import pymbolic as pmbl
from functools import partial


def main():

    print("*************************")
    print("* Setting up...")
    print("*************************")

    dim = 2

    # download precomputation results for the 2D Laplace kernel
    download_table = True
    table_filename = "nft_laplace2d.hdf5"
    root_table_source_extent = 2

    print("Using table cache:", table_filename)

    q_order = 9   # quadrature order
    n_levels = 6  # 2^(n_levels-1) subintervals in 1D

    use_multilevel_table = False

    adaptive_mesh = False
    n_refinement_loops = 100
    refined_n_cells = 2000
    rratio_top = 0.2
    rratio_bot = 0.5

    dtype = np.float64

    m_order = 20  # multipole order
    force_direct_evaluation = False

    print("Multipole order =", m_order)

    alpha = 160

    x = pmbl.var("x")
    y = pmbl.var("y")
    expp = pmbl.var("exp")

    norm2 = x ** 2 + y ** 2
    source_expr = -(4 * alpha ** 2 * norm2 - 4 * alpha) * expp(-alpha * norm2)
    solu_expr = expp(-alpha * norm2)

    logger.info("Source expr: " + str(source_expr))
    logger.info("Solu expr: " + str(solu_expr))

    # bounding box
    a = -0.5
    b = 0.5
    root_table_source_extent = 2

    ctx = cl.create_some_context()
    queue = cl.CommandQueue(ctx)

    source_eval = Eval(dim, source_expr, [x, y])

    # {{{ generate quad points

    import volumential.meshgen as mg

    # Show meshgen info
    mg.greet()

    mesh = mg.MeshGen2D(q_order, n_levels, a, b, queue=queue)
    if not adaptive_mesh:
        mesh.print_info()
        q_points = mesh.get_q_points()
        q_weights = mesh.get_q_weights()

    else:
        iloop = -1
        while mesh.n_active_cells() < refined_n_cells:
            iloop += 1
            crtr = np.abs(
                source_eval(mesh.get_cell_centers)
                * mesh.get_cell_measures)
            mesh.update_mesh(crtr, rratio_top, rratio_bot)
            if iloop > n_refinement_loops:
                print("Max number of refinement loops reached.")
                break

        mesh.print_info()
        q_points = mesh.get_q_points()
        q_weights = mesh.get_q_weights()

    assert len(q_points) == len(q_weights)
    assert q_points.shape[1] == dim

    q_points = np.ascontiguousarray(np.transpose(q_points))

    from pytools.obj_array import make_obj_array

    q_points = make_obj_array(
        [cl.array.to_device(queue, q_points[i]) for i in range(dim)])

    q_weights = cl.array.to_device(queue, q_weights)
    # q_radii = cl.array.to_device(queue, q_radii)

    # }}}

    # {{{ discretize the source field

    source_vals = cl.array.to_device(
        queue, source_eval(queue, np.array([coords.get() for coords in q_points]))
    )

    # particle_weigt = source_val * q_weight

    # }}} End discretize the source field

    # {{{ build tree and traversals

    from boxtree.tools import AXIS_NAMES

    axis_names = AXIS_NAMES[:dim]

    from pytools import single_valued

    coord_dtype = single_valued(coord.dtype for coord in q_points)
    from boxtree.bounding_box import make_bounding_box_dtype

    bbox_type, _ = make_bounding_box_dtype(ctx.devices[0], dim, coord_dtype)
    bbox = np.empty(1, bbox_type)
    for ax in axis_names:
        bbox["min_" + ax] = a
        bbox["max_" + ax] = b

    # tune max_particles_in_box to reconstruct the mesh
    # TODO: use points from FieldPlotter are used as target points for better
    # visuals
    from boxtree import TreeBuilder

    tb = TreeBuilder(ctx)
    tree, _ = tb(
        queue,
        particles=q_points,
        targets=q_points,
        bbox=bbox,
        max_particles_in_box=q_order ** 2 * 4 - 1,
        kind="adaptive-level-restricted",
    )

    bbox2 = np.array([[a, b], [a, b]])
    tree2, _ = tb(
        queue,
        particles=q_points,
        targets=q_points,
        bbox=bbox2,
        max_particles_in_box=q_order ** 2 * 4 - 1,
        kind="adaptive-level-restricted",
    )

    from boxtree.traversal import FMMTraversalBuilder

    tg = FMMTraversalBuilder(ctx)
    trav, _ = tg(queue, tree)

    # }}} End build tree and traversals

    # {{{ build near field potential table

    from volumential.table_manager import NearFieldInteractionTableManager
    import os

    if download_table and (not os.path.isfile(table_filename)):
        import json
        with open("table_urls.json", 'r') as fp:
            urls = json.load(fp)

        print("Downloading table from %s" % urls['Laplace2D'])
        import subprocess
        subprocess.call(["wget", "-q", urls['Laplace2D'], table_filename])

    tm = NearFieldInteractionTableManager(
        table_filename, root_extent=root_table_source_extent,
        queue=queue)

    if use_multilevel_table:
        assert (
            abs(
                int((b - a)
                  / root_table_source_extent) * root_table_source_extent
                - (b - a))
            < 1e-15)
        nftable = []
        for l in range(0, tree.nlevels + 1):
            print("Getting table at level", l)
            tb, _ = tm.get_table(
                dim,
                "Laplace",
                q_order,
                source_box_level=l,
                compute_method="DrosteSum",
                queue=queue,
                n_brick_quad_poitns=100,
                adaptive_level=False,
                use_symmetry=True,
                alpha=0.1,
                nlevels=15,
            )
            nftable.append(tb)

        print("Using table list of length", len(nftable))

    else:
        nftable, _ = tm.get_table(
            dim,
            "Laplace",
            q_order,
            force_recompute=False,
            compute_method="DrosteSum",
            queue=queue,
            n_brick_quad_poitns=100,
            adaptive_level=False,
            use_symmetry=True,
            alpha=0.1,
            nlevels=15,
        )

    # }}} End build near field potential table

    # {{{ sumpy expansion for laplace kernel

    from sumpy.expansion import DefaultExpansionFactory
    from sumpy.kernel import LaplaceKernel

    knl = LaplaceKernel(dim)
    out_kernels = [knl]

    expn_factory = DefaultExpansionFactory()
    local_expn_class = expn_factory.get_local_expansion_class(knl)
    mpole_expn_class = expn_factory.get_multipole_expansion_class(knl)

    exclude_self = True

    from volumential.expansion_wrangler_fpnd import (
            FPNDExpansionWranglerCodeContainer,
            FPNDExpansionWrangler
            )

    wcc = FPNDExpansionWranglerCodeContainer(
        ctx,
        partial(mpole_expn_class, knl),
        partial(local_expn_class, knl),
        out_kernels,
        exclude_self=exclude_self,
    )

    if exclude_self:
        target_to_source = np.arange(tree.ntargets, dtype=np.int32)
        self_extra_kwargs = {"target_to_source": target_to_source}
    else:
        self_extra_kwargs = {}

    wrangler = FPNDExpansionWrangler(
        code_container=wcc,
        queue=queue,
        tree=tree,
        near_field_table=nftable,
        dtype=dtype,
        fmm_level_to_order=lambda kernel, kernel_args, tree, lev: m_order,
        quad_order=q_order,
        self_extra_kwargs=self_extra_kwargs,
    )

    # }}} End sumpy expansion for laplace kernel

    print("*************************")
    print("* Performing FMM ...")
    print("*************************")

    # {{{ conduct fmm computation

    from volumential.volume_fmm import drive_volume_fmm

    import time
    queue.finish()

    t0 = time.clock()

    pot, = drive_volume_fmm(
        trav,
        wrangler,
        source_vals * q_weights,
        source_vals,
        direct_evaluation=force_direct_evaluation,
    )
    queue.finish()

    t1 = time.clock()

    print("Finished in %.2f seconds." % (t1 - t0))
    print("(%e points per second)" % (
        len(q_weights) / (t1 - t0)
        ))

    # }}} End conduct fmm computation

    print("*************************")
    print("* Postprocessing ...")
    print("*************************")

    # {{{ postprocess and plot

    # print(pot)

    solu_eval = Eval(dim, solu_expr, [x, y])

    x = q_points[0].get()
    y = q_points[1].get()
    ze = solu_eval(queue, np.array([x, y]))
    zs = pot.get()

    print_error = True
    if print_error:
        err = np.max(np.abs(ze - zs))
        print("Error =", err)

    # Interpolated surface
    if 0:
        h = 0.005
        out_x = np.arange(a, b + h, h)
        out_y = np.arange(a, b + h, h)
        oxx, oyy = np.meshgrid(out_x, out_y)
        out_targets = make_obj_array(
            [
                cl.array.to_device(queue, oxx.flatten()),
                cl.array.to_device(queue, oyy.flatten()),
            ]
        )

        from volumential.volume_fmm import interpolate_volume_potential

        # src = source_field([q.get() for q in q_points])
        # src = cl.array.to_device(queue, src)
        interp_pot = interpolate_volume_potential(out_targets, trav, wrangler, pot)
        opot = interp_pot.get()

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        plt3d = plt.figure()
        ax = Axes3D(plt3d)  # noqa
        surf = ax.plot_surface(oxx, oyy, opot.reshape(oxx.shape))  # noqa
        # ax.scatter(x, y, src.get())
        # ax.set_zlim(-0.25, 0.25)

        plt.draw()
        plt.show()

    # Boxtree
    if 0:
        import matplotlib.pyplot as plt

        if dim == 2:
            # plt.plot(q_points[0].get(), q_points[1].get(), ".")
            pass

        from boxtree.visualization import TreePlotter

        plotter = TreePlotter(tree.get(queue=queue))
        plotter.draw_tree(fill=False, edgecolor="black")
        # plotter.draw_box_numbers()
        plotter.set_bounding_box()
        plt.gca().set_aspect("equal")

        plt.draw()
        # plt.show()
        plt.savefig("tree.png")

    # Direct p2p
    if 0:
        print("Performing P2P")
        pot_direct, = drive_volume_fmm(
            trav, wrangler, source_vals * q_weights,
            source_vals, direct_evaluation=True)
        zds = pot_direct.get()
        zs = pot.get()

        print("P2P-FMM diff =", np.max(np.abs(zs - zds)))

        print("P2P Error =", np.max(np.abs(ze - zds)))

        """
        import matplotlib.pyplot as plt
        import matplotlib.cm as cm
        x = q_points[0].get()
        y = q_points[1].get()
        plt.scatter(x, y, c=np.log(abs(zs-zds)) / np.log(10), cmap=cm.jet)
        plt.colorbar()

        plt.xlabel("Multipole order = " + str(m_order))

        plt.draw()
        plt.show()
        """

    # Scatter plot
    if 0:
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        x = q_points[0].get()
        y = q_points[1].get()
        ze = solu_eval(queue, np.array([x, y]))
        zs = pot.get()

        plt3d = plt.figure()
        ax = Axes3D(plt3d)
        ax.scatter(x, y, zs, s=1)
        # ax.scatter(x, y, source_field([q.get() for q in q_points]), s=1)
        # import matplotlib.cm as cm

        # ax.scatter(x, y, zs, c=np.log(abs(zs-zds)), cmap=cm.jet)
        # plt.gca().set_aspect("equal")

        # ax.set_xlim3d([-1, 1])
        # ax.set_ylim3d([-1, 1])
        # ax.set_zlim3d([np.min(z), np.max(z)])
        # ax.set_zlim3d([-0.002, 0.00])

        plt.draw()
        plt.show()
        # plt.savefig("exact.png")

    # }}} End postprocess and plot


if __name__ == '__main__':
    main()


# vim: filetype=pyopencl:foldmethod=marker



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__doc__ = """
Mesh generation

.. autoclass:: MeshGenBase
   :members:
.. autoclass:: MeshGen2D
   :members:
.. autoclass:: MeshGen3D
   :members:
"""

import logging
import numpy as np
import pyopencl as cl
from pytools.obj_array import make_obj_array

logger = logging.getLogger(__name__)

provider = None

# {{{ meshgen Python provider


class MeshGenBase(object):
    """Base class for Meshgen via BoxTree.
    The interface is similar to the Meshgen via Deal.II, except that
    the arguments a and b can also be of higher dimensions to allow
    added flexibility on choosing the bounding box.

    This base class cannot be used directly since the boxtree is not
    built in the constructor until dimension_specific_setup() is
    provided.

    In addition to the common capabilities of the deal.II implementation,
    this class also implements native CL getters like get_q_points_dev()
    that play well with the other libraries like boxtree.
    """

    def __init__(self, degree, nlevels, a=-1, b=1, queue=None):
        assert degree > 0
        assert nlevels > 0
        self.degree = degree
        self.quadrature_formula = LegendreGaussQuadrature(degree - 1)
        self.nlevels = nlevels

        self.bound_a = np.array([a]).flatten()
        self.bound_b = np.array([b]).flatten()
        assert len(self.bound_a) == len(self.bound_b)
        assert np.all(self.bound_a < self.bound_b)

        self.dim = len(self.bound_a)
        self.root_vertex = self.bound_a
        self.root_extent = np.max(self.bound_b - self.bound_a)

        if queue is None:
            ctx = cl.create_some_context()
            self.queue = cl.CommandQueue(ctx)
        else:
            self.queue = queue

        # plug in dimension-specific details
        self.dimension_specific_setup()
        self.n_q_points = self.degree ** self.dim

        self.boxtree = BoxTree()
        self.boxtree.generate_uniform_boxtree(
            self.queue,
            nlevels=self.nlevels,
            root_extent=self.root_extent,
            root_vertex=self.root_vertex,
        )
        self.quadrature = QuadratureOnBoxTree(
            self.boxtree, self.quadrature_formula
        )

    def dimension_specific_setup(self):
        pass

    def get_q_points_dev(self):
        return self.quadrature.get_q_points(self.queue)

    def get_q_points(self):
        q_points_dev = self.get_q_points_dev()
        n_all_q_points = len(q_points_dev[0])
        q_points = np.zeros((n_all_q_points, self.dim))
        for d in range(self.dim):
            q_points[:, d] = q_points_dev[d].get(self.queue)
        return q_points

    def get_q_weights_dev(self):
        return self.quadrature.get_q_weights(self.queue)

    def get_q_weights(self):
        q_weights_dev = self.get_q_weights_dev()
        return q_weights_dev.get(self.queue)

    def get_cell_measures_dev(self):
        return self.quadrature.get_cell_measures(self.queue)

    def get_cell_measures(self):
        cell_measures_dev = self.get_cell_measures_dev()
        return cell_measures_dev.get(self.queue)

    def get_cell_centers_dev(self):
        return self.quadrature.get_cell_centers(self.queue)

    def get_cell_centers(self):
        cell_centers_dev = self.get_cell_centers_dev()
        n_active_cells = self.n_active_cells()
        cell_centers = np.zeros((n_active_cells, self.dim))
        for d in range(self.dim):
            cell_centers[:, d] = cell_centers_dev[d].get(self.queue)
        return cell_centers

    def n_cells(self):
        """Note that this value can be larger than the actual number
        of cells used in the boxtree. It mainly serves as the bound for
        iterators on cells.
        """
        return self.boxtree.nboxes

    def n_active_cells(self):
        return self.boxtree.n_active_boxes

    def update_mesh(
        self, criteria, top_fraction_of_cells, bottom_fraction_of_cells
    ):
        # TODO
        raise NotImplementedError

    def print_info(self, logging_func=logger.info):
        logging_func("Number of cells: " + str(self.n_cells()))
        logging_func("Number of active cells: " + str(self.n_active_cells()))
        logging_func("Number of quad points per cell: "
                + str(self.n_q_points))

    def generate_gmsh(self, filename):
        """
        # TODO
        Write the active boxes as a gmsh file.
        The file format specifications can be found at:
        http://gmsh.info/doc/texinfo/gmsh.html#MSH-ASCII-file-format
        """
        raise NotImplementedError


# }}} End meshgen Python provider

try:
    logger.info("Trying to find a mesh generator..")
    import volumential.meshgen_dealii  # noqa: F401

    provider = "meshgen_dealii"

except ImportError as e:
    logger.debug(repr(e))
    logger.warn("Meshgen via Deal.II is not present or unusable.")

    try:
        logger.info("Trying out BoxTree.TreeInteractiveBuild interface.")
        import boxtree.tree_interactive_build  # noqa: F401
        from modepy import LegendreGaussQuadrature

        provider = "meshgen_boxtree"

    except ImportError as ee:
        logger.debug(repr(ee))
        logger.warn("Meshgen via BoxTree is not present or unusable.")
        raise RuntimeError("Cannot find a usable Meshgen implementation.")

    else:
        # {{{ Meshgen via BoxTree
        logger.info("Using Meshgen via BoxTree interface.")
        from boxtree.tree_interactive_build import BoxTree, QuadratureOnBoxTree

        def greet():
            return "Hello from Meshgen via BoxTree!"

        def make_uniform_cubic_grid(degree, nlevels=1, dim=2, queue=None, **kwargs):
            """Uniform cubic grid in [-1,1]^dim.
            This function provides backward compatibility with meshgen_dealii.
            """
            if queue is None:
                ctx = cl.create_some_context()
                queue = cl.CommandQueue(ctx)

            # For meshgen_dealii compatibility
            if "level" in kwargs:
                nlevels = kwargs["level"]

            tree = BoxTree()
            tree.generate_uniform_boxtree(
                queue, nlevels=nlevels, root_extent=2, root_vertex=np.zeros(dim) - 1
            )
            quad_rule = LegendreGaussQuadrature(degree - 1)
            quad = QuadratureOnBoxTree(tree, quad_rule)
            q_weights = quad.get_q_weights(queue).get(queue)
            q_points_dev = quad.get_q_points(queue)
            n_all_q_points = len(q_weights)
            q_points = np.zeros((n_all_q_points, dim))
            for d in range(dim):
                q_points[:, d] = q_points_dev[d].get(queue)

            # Adding a placeholder for deprecated point radii
            return (q_points, q_weights, None)

        class MeshGen1D(MeshGenBase):
            """Meshgen in 1D
            """

            def dimension_specific_setup(self):
                assert self.dim == 1

        class MeshGen2D(MeshGenBase):
            """Meshgen in 2D
            """

            def dimension_specific_setup(self):
                if self.dim == 1:
                    # allow passing scalar values of a and b to the constructor
                    self.dim = 2
                    self.root_vertex = np.zeros(self.dim) + self.bound_a
                else:
                    assert self.dim == 2

        class MeshGen3D(MeshGenBase):
            """Meshgen in 3D
            """

            def dimension_specific_setup(self):
                if self.dim == 1:
                    # allow passing scalar values of a and b to the constructor
                    self.dim = 3
                    self.root_vertex = np.zeros(self.dim) + self.bound_a
                else:
                    assert self.dim == 3

        # }}} End Meshgen via BoxTree

else:
    # noexcept on importing meshgen_dealii
    logger.info("Using Meshgen via Deal.II interface.")
    from volumential.meshgen_dealii import (  # noqa: F401
        greet,
        MeshGen2D,
        MeshGen3D,
    )

    def make_uniform_cubic_grid(degree, level, dim, queue=None):
        from volumential.meshgen_dealii import make_uniform_cubic_grid as _mucg
        return _mucg(degree, level, dim)


# {{{ mesh utils

def build_geometry_info(ctx, queue, dim, q_order, mesh,
                             bbox=None, a=None, b=None):
    """Build tree, traversal and other geo info for FMM computation,
    given the box mesh over/encompassing the domain.

    The bouding box can be specified in one of two ways:
    1. via scalars a, b, dim-homogeneous ([a, b]^dim)
    2. via bbox (e.g. np.array([[a1, b1], [a2, b2], [a3, b3]]))
    """

    if dim == 1:
        if not isinstance(mesh, MeshGen1D):
            raise ValueError()

    if dim == 2:
        if not isinstance(mesh, MeshGen2D):
            raise ValueError()

    elif dim == 3:
        if not isinstance(mesh, MeshGen3D):
            raise ValueError()

    else:
        raise ValueError("only supports 1 <= dim <= 3")

    q_points = mesh.get_q_points()
    q_weights = mesh.get_q_weights()
    q_points_org = q_points  # noqa: F841
    q_points = np.ascontiguousarray(np.transpose(q_points))

    q_points = make_obj_array(
            [cl.array.to_device(queue, q_points[i])
                for i in range(dim)])
    q_weights = cl.array.to_device(queue, q_weights)

    if bbox is None:
        assert np.isscalar(a) and np.isscalar(b)
        bbox = np.array([[a, b], ] * dim)

    from boxtree import TreeBuilder
    tb = TreeBuilder(ctx)

    tree, _ = tb(
        queue,
        particles=q_points,
        targets=q_points,
        bbox=bbox,
        max_particles_in_box=q_order**dim * (2**dim) - 1,
        kind="adaptive-level-restricted",
    )

    from boxtree.traversal import FMMTraversalBuilder
    tg = FMMTraversalBuilder(ctx)
    trav, _ = tg(queue, tree)

    return q_points, q_weights, tree, trav

# }}} End mesh utils

# vim: ft=pyopencl



from __future__ import absolute_import, division, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import os
from pytools.persistent_dict import WriteOncePersistentDict

from volumential.singular_integral_2d import box_quad
from volumential.version import VERSION_TEXT
from volumential.table_manager import NearFieldInteractionTableManager  # noqa: F401
from volumential.nearfield_potential_table import (  # noqa: F401
    NearFieldInteractionTable,
)

volumential_version = VERSION_TEXT

__all__ = ["volumential_version", "box_quad", "nearfield_potential_table"]

__doc__ = """
:mod:`volumential` can compute 2/3D volume potentials using FMM.
"""

code_cache = WriteOncePersistentDict("volumential-code-cache-v0-"+VERSION_TEXT)

# {{{ optimization control

OPT_ENABLED = True

OPT_ENABLED = "VOLUMENTIAL_NO_OPT" not in os.environ


def set_optimization_enabled(flag):
    """Set whether the :mod:`loopy` kernels should be optimized."""
    global OPT_ENABLED
    OPT_ENABLED = flag

# }}}

# {{{ cache control


CACHING_ENABLED = True

CACHING_ENABLED = (
    "VOLUMENTIAL_NO_CACHE" not in os.environ
    and "CG_NO_CACHE" not in os.environ)


def set_caching_enabled(flag):
    """Set whether :mod:`loopy` is allowed to use disk caching for its various
    code generation stages.
    """
    global CACHING_ENABLED
    CACHING_ENABLED = flag


class CacheMode(object):
    """A context manager for setting whether :mod:`volumential` is allowed to use
    disk caches.
    """

    def __init__(self, new_flag):
        self.new_flag = new_flag

    def __enter__(self):
        global CACHING_ENABLED
        self.previous_mode = CACHING_ENABLED
        CACHING_ENABLED = self.new_flag

    def __exit__(self, exc_type, exc_val, exc_tb):
        global CACHING_ENABLED
        CACHING_ENABLED = self.previous_mode
        del self.previous_mode

# }}}

# vim: filetype=pyopencl:fdm=marker



from __future__ import division, absolute_import, print_function

__copyright__ = "Copyright (C) 2017 - 2018 Xiaoyu Wei"

__doc__ = """
.. autoclass:: NearFieldInteractionTableManager
   :members:
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import h5py as hdf
import numpy as np

import volumential as vm

# from sumpy.expansion.local import VolumeTaylorLocalExpansion
# from sumpy.expansion.multipole import VolumeTaylorMultipoleExpansion
# from sumpy.kernel import LaplaceKernel
from volumential.nearfield_potential_table import NearFieldInteractionTable

logger = logging.getLogger(__name__)

# {{{ constant sumpy kernel

from sumpy.kernel import ExpressionKernel


class ConstantKernel(ExpressionKernel):
    init_arg_names = ("dim",)

    def __init__(self, dim=None):

        expr = 1
        scaling = 1

        super(ConstantKernel, self).__init__(
            dim, expression=expr,
            global_scaling_const=scaling, is_complex_valued=False
        )

    has_efficient_scale_adjustment = True

    def adjust_for_kernel_scaling(self, expr, rscale, nderivatives):
        return expr / rscale

    def __getinitargs__(self):
        return (self.dim,)

    def __repr__(self):
        return "CstKnl%dD" % self.dim

    mapper_method = "map_expression_kernel"


# }}} End constant sumpy kernel

# {{{ table dataset manager class


class NearFieldInteractionTableManager(object):
    """
    A class that manages near field interaction table computation and
    storage.

    Tables are stored under 'Dimension/KernelName/QuadOrder/BoxLevel/dataset_name'
    e.g., '2D/Laplace/Order_1/Level_0/data'

    Only one table manager can exist for a dataset file with write access.
    The access can be controlled with the read_only argument. By default,
    the constructor tries to open the dataset with write access, and falls
    back to read-only if that fails.
    """

    def __init__(self, dataset_filename="nft.hdf5",
            root_extent=1, dtype=np.float64,
            read_only='auto', **kwargs):
        """Constructor.
        """
        self.dtype = dtype

        self.filename = dataset_filename
        self.root_extent = root_extent

        if read_only == 'auto':
            try:
                self.datafile = hdf.File(self.filename, "a")
            except (IOError, OSError) as e:
                from warnings import warn
                warn("Trying to open in read/write mode failed: %s" % str(e))
                warn("Opening table dataset %s in read-only mode." % self.filename)
                self.datafile = hdf.File(self.filename, "r")
        elif read_only:
            self.datafile = hdf.File(self.filename, "r")
        else:
            # Read/write if exists, create otherwise
            self.datafile = hdf.File(self.filename, "a")

        self.table_extra_kwargs = kwargs

        # If the file exists, it must be for the same root_extent
        if "root_extent" in self.datafile.attrs:
            if not abs(
                    self.datafile.attrs["root_extent"] - self.root_extent
                    ) < 1e-15:
                raise RuntimeError(
                    "The table cache file "
                    + self.filename
                    + " was built with root_extent = "
                    + str(self.datafile.attrs["root_extent"])
                    + ", which is different from the requested value "
                    + str(self.root_extent)
                )

        self.supported_kernels = [
            "Laplace",
            "Laplace-Dx",
            "Laplace-Dy",
            "Laplace-Dz",
            "Constant",
            "Yukawa",
            "Yukawa-Dx",
            "Yukawa-Dy",
            "Cahn-Hilliard",
            "Cahn-Hilliard-Laplacian",
            "Cahn-Hilliard-Dx",
            "Cahn-Hilliard-Dy",
            "Cahn-Hilliard-Laplacian-Dx",
            "Cahn-Hilliard-Laplacian-Dy",
        ]

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.datafile.__exit__(exception_type, exception_value, traceback)

    def get_table(
        self,
        dim,
        kernel_type,
        q_order,
        source_box_level=0,
        force_recompute=False,
        compute_method=None,
        queue=None,
        **kwargs
    ):
        """Primary user interface. Get the specified table regardless of how.
        In the case of a cache miss or a forced re-computation, the method
        specified in the compute_method will be used.
        """

        dim = int(dim)
        assert dim >= 1

        if compute_method is None:
            compute_method = "Transform"

        is_recomputed = False

        q_order = int(q_order)
        assert q_order >= 1

        source_box_level = int(source_box_level)
        assert source_box_level >= 0

        if str(dim) + "D" not in self.datafile:
            self.datafile.create_group(str(dim) + "D")

        grp = self.datafile[str(dim) + "D"]

        if kernel_type not in grp:
            grp.create_group(kernel_type)

        grp = grp[kernel_type]

        if "Order_" + str(q_order) not in grp:
            grp.create_group("Order_" + str(q_order))

        grp = grp["Order_" + str(q_order)]

        if "Level_" + str(source_box_level) not in grp:
            logger.info("Table cache missing. Invoking fresh computation.")
            is_recomputed = True
            grp.create_group("Level_" + str(source_box_level))
            table = self.compute_and_update_table(
                dim,
                kernel_type,
                q_order,
                source_box_level,
                compute_method,
                queue=queue,
                **kwargs
            )

        elif force_recompute:
            logger.info("Invoking fresh computation since force_recompute is set")
            is_recomputed = True
            table = self.compute_and_update_table(
                dim,
                kernel_type,
                q_order,
                source_box_level,
                compute_method,
                queue=queue,
                **kwargs
            )

        else:
            try:

                table = self.load_saved_table(
                    dim,
                    kernel_type,
                    q_order,
                    source_box_level,
                    compute_method,
                    **kwargs
                )

            except KeyError:

                import traceback

                logger.debug(traceback.format_exc())

                logger.info("Recomputing due to table data corruption.")
                is_recomputed = True
                table = self.compute_and_update_table(
                    dim,
                    kernel_type,
                    q_order,
                    source_box_level,
                    compute_method,
                    queue=queue,
                    **kwargs
                )

            # Ensure loaded table matches requirements specified in kwargs
            for kkey, kval in kwargs.items():
                if kval is not None:
                    try:
                        tbval = getattr(table, kkey)
                        if isinstance(kval, (int, str)):
                            if not kval == tbval:
                                from warnings import warn

                                warn(
                                    "Table data loaded with a different value "
                                    + kkey
                                    + " = "
                                    + str(tbval)
                                    + " (expected "
                                    + str(kval)
                                    + ")"
                                )
                        else:
                            assert isinstance(kval, (float, complex))
                            tbval = kval.__class__(tbval)
                            if not abs(kval - tbval) < 1e-12:
                                from warnings import warn

                                warn(
                                    "Table data loaded with a different value "
                                    + kkey
                                    + " = "
                                    + str(tbval)
                                    + " (expected "
                                    + str(kval)
                                    + ")"
                                )

                    except AttributeError as e:
                        strict_loading = False
                        if "debug" in kwargs:
                            if "strict_loading" in kwargs["debug"]:
                                strict_loading = kwargs["debug"]["strict_loading"]

                        if strict_loading:
                            from warnings import warn

                            warn(
                                "Consistency is not fully ensured "
                                "(kwarg specified but cannot be loaded). "
                                "NOTE: this is most likely due to non-standard "
                                "arguements being passed, since only "
                                "(int, float, complex, str) "
                                "are stored in the cache. "
                                "Also, some parameters related to method for "
                                "table building are not critical for "
                                "consistency."
                            )
                            print(e)

        return table, is_recomputed

    def load_saved_table(
        self,
        dim,
        kernel_type,
        q_order,
        source_box_level=0,
        compute_method=None,
        **kwargs
    ):
        """Load a table saved in the hdf5 file.
        """

        q_order = int(q_order)
        assert q_order >= 1

        source_box_level = int(source_box_level)
        assert source_box_level >= 0

        # Check data table integrity
        assert str(dim) + "D" in self.datafile
        assert kernel_type in self.datafile[str(dim) + "D"]
        assert "Order_" + str(q_order) in self.datafile[str(dim) + "D"][kernel_type]
        assert (
            "Level_" + str(source_box_level)
            in self.datafile[str(dim) + "D"][kernel_type]["Order_" + str(q_order)]
        )

        grp = self.datafile[str(dim) + "D"][kernel_type]["Order_" + str(q_order)][
            "Level_" + str(source_box_level)
        ]

        assert dim == grp.attrs["dim"]
        assert q_order == grp.attrs["quad_order"]

        if compute_method == "Transform":
            if 'knl_func' not in kwargs:
                knl_func = self.get_kernel_function(dim, kernel_type, **kwargs)
            else:
                knl_func = kwargs['knl_func']
            sumpy_knl = None
        elif compute_method == "DrosteSum":
            knl_func = None
            if 'sumpy_knl' not in kwargs:
                sumpy_knl = self.get_sumpy_kernel(dim, kernel_type)
            else:
                sumpy_knl = kwargs['sumpy_knl']
        else:
            from warnings import warn

            warn("Unsupported compute_method: ", compute_method)
            knl_func = None
            sumpy_knl = None

        table = NearFieldInteractionTable(
            quad_order=q_order,
            dim=dim,
            dtype=self.dtype,
            build_method=compute_method,
            kernel_func=knl_func,
            kernel_type=self.get_kernel_function_type(dim, kernel_type),
            sumpy_kernel=sumpy_knl,
            source_box_extent=self.root_extent * (2 ** (-source_box_level)),
            **self.table_extra_kwargs
        )

        assert abs(table.source_box_extent - grp.attrs["source_box_extent"]) < 1e-15
        assert source_box_level == grp.attrs["source_box_level"]

        # Load data
        table.q_points[...] = grp["q_points"]
        table.data[...] = grp["data"]
        if 'mode_normalizers' in grp:
            table.mode_normalizers[...] = grp["mode_normalizers"]
        if 'kernel_exterior_normalizers' in grp:
            table.kernel_exterior_normalizers[...] = \
                    grp["kernel_exterior_normalizers"]

        tmp_case_vecs = np.array(table.interaction_case_vecs)
        tmp_case_vecs[...] = grp["interaction_case_vecs"]
        table.interaction_case_vecs = [list(vec) for vec in tmp_case_vecs]

        table.case_indices[...] = grp["case_indices"]

        assert table.n_q_points == grp.attrs["n_q_points"]
        assert table.n_pairs == grp.attrs["n_pairs"]
        assert table.quad_order == grp.attrs["quad_order"]

        base = grp.attrs["case_encoding_base"]
        shift = grp.attrs["case_encoding_shift"]

        def case_encode(case_vec):
            table_id = 0
            for l in case_vec:
                table_id = table_id * base + (l + shift)
            return int(table_id)

        table.case_encode = case_encode

        # load extra kwargs
        for atkey, atval in grp.attrs.items():
            setattr(table, atkey, atval)

        table.is_built = True

        return table

    def get_kernel_function(self, dim, kernel_type, **kwargs):
        """Kernel function is needed for building the table. This function
        provides support for some kernels such that the user can build and
        use the table without explicitly providing such information.
        """
        if kernel_type == "Laplace":
            knl_func = vm.nearfield_potential_table.get_laplace(dim)
        elif kernel_type == "Constant":
            knl_func = vm.nearfield_potential_table.constant_one
        elif kernel_type == "Cahn-Hilliard":
            knl_func = vm.nearfield_potential_table.get_cahn_hilliard(
                dim, kwargs["b"], kwargs["c"]
            )
        elif kernel_type in self.supported_kernels:
            knl = self.get_sumpy_kernel(dim, kernel_type)
            knl_func = vm.nearfield_potential_table.sumpy_kernel_to_lambda(knl)
        else:
            raise NotImplementedError("Kernel type not supported.")

        return knl_func

    def get_sumpy_kernel(self, dim, kernel_type):
        """Sumpy (symbolic) version of the kernel.
        """

        if kernel_type == "Laplace":
            from sumpy.kernel import LaplaceKernel

            return LaplaceKernel(dim)

        if kernel_type == "Laplace-Dx":
            from sumpy.kernel import LaplaceKernel, AxisTargetDerivative

            return AxisTargetDerivative(0, LaplaceKernel(dim))

        if kernel_type == "Laplace-Dy":
            from sumpy.kernel import LaplaceKernel, AxisTargetDerivative

            return AxisTargetDerivative(1, LaplaceKernel(dim))

        if kernel_type == "Laplace-Dz":
            from sumpy.kernel import LaplaceKernel, AxisTargetDerivative

            assert dim >= 3
            return AxisTargetDerivative(2, LaplaceKernel(dim))

        elif kernel_type == "Constant":
            return ConstantKernel(dim)

        elif kernel_type == "Yukawa":
            from sumpy.kernel import YukawaKernel

            return YukawaKernel(dim)

        elif kernel_type == "Yukawa-Dx":
            from sumpy.kernel import YukawaKernel, AxisTargetDerivative

            return AxisTargetDerivative(0, YukawaKernel(dim))

        elif kernel_type == "Yukawa-Dy":
            from sumpy.kernel import YukawaKernel, AxisTargetDerivative

            return AxisTargetDerivative(1, YukawaKernel(dim))

        elif kernel_type == "Cahn-Hilliard":
            from sumpy.kernel import FactorizedBiharmonicKernel

            return FactorizedBiharmonicKernel(dim)

        elif kernel_type == "Cahn-Hilliard-Laplacian":
            from sumpy.kernel import (
                FactorizedBiharmonicKernel,
                LaplacianTargetDerivative,
            )

            return LaplacianTargetDerivative(FactorizedBiharmonicKernel(dim))

        elif kernel_type == "Cahn-Hilliard-Dx":
            from sumpy.kernel import FactorizedBiharmonicKernel, AxisTargetDerivative

            return AxisTargetDerivative(0, FactorizedBiharmonicKernel(dim))

        elif kernel_type == "Cahn-Hilliard-Laplacian-Dx":
            from sumpy.kernel import (
                FactorizedBiharmonicKernel,
                LaplacianTargetDerivative,
            )
            from sumpy.kernel import AxisTargetDerivative

            return AxisTargetDerivative(
                0, LaplacianTargetDerivative(FactorizedBiharmonicKernel(dim))
            )

        elif kernel_type == "Cahn-Hilliard-Laplacian-Dy":
            from sumpy.kernel import (
                FactorizedBiharmonicKernel,
                LaplacianTargetDerivative,
            )
            from sumpy.kernel import AxisTargetDerivative

            return AxisTargetDerivative(
                1, LaplacianTargetDerivative(FactorizedBiharmonicKernel(dim))
            )

        elif kernel_type == "Cahn-Hilliard-Dy":
            from sumpy.kernel import FactorizedBiharmonicKernel, AxisTargetDerivative

            return AxisTargetDerivative(1, FactorizedBiharmonicKernel(dim))

        elif kernel_type in self.supported_kernels:
            return None

        else:
            raise NotImplementedError("Kernel type not supported.")

    def get_kernel_function_type(self, dim, kernel_type):
        """Determines how and to what extend the table data can be rescaled
        and reused.
        """

        if kernel_type == "Laplace":
            if dim == 2:
                return "log"
            elif dim >= 3:
                return "inv_power"
            else:
                raise NotImplementedError("Kernel scaling not supported")

        elif kernel_type == "Constant":
            if dim >= 1 and dim <= 3:
                return "const"
            else:
                raise NotImplementedError("Kernel scaling not supported")

        else:
            return None

    def update_dataset(self, group, dataset_name, data_array):
        """Update stored data.
        """
        if data_array is None:
            logger.debug("No data to save for %s" % dataset_name)
            return

        data_array = np.array(data_array)

        if dataset_name not in group:
            dset = group.create_dataset(
                dataset_name, data_array.shape, dtype=data_array.dtype
            )
        else:
            dset = group[dataset_name]

        dset[...] = data_array

    def compute_and_update_table(
        self,
        dim,
        kernel_type,
        q_order,
        source_box_level=0,
        compute_method=None,
        cl_ctx=None,
        queue=None,
        **kwargs
    ):
        """Performs the precomputation and stores the results.
        """

        if compute_method is None:
            logger.debug("Using default compute_method (Transform)")
            compute_method = "Transform"

        self.datafile.attrs["root_extent"] = self.root_extent

        q_order = int(q_order)
        assert q_order >= 1

        assert str(dim) + "D" in self.datafile
        assert kernel_type in self.datafile[str(dim) + "D"]
        assert "Order_" + str(q_order) in self.datafile[str(dim) + "D"][kernel_type]

        if compute_method == "Transform":
            if 'knl_func' not in kwargs:
                knl_func = self.get_kernel_function(dim, kernel_type, **kwargs)
            else:
                knl_func = kwargs['knl_func']
            sumpy_knl = None
        elif compute_method == "DrosteSum":
            knl_func = None
            if 'sumpy_knl' not in kwargs:
                sumpy_knl = self.get_sumpy_kernel(dim, kernel_type)
            else:
                sumpy_knl = kwargs['sumpy_knl']
        else:
            raise NotImplementedError("Unsupported compute_method.")

        knl_type = self.get_kernel_function_type(dim, kernel_type)

        # compute table
        logger.debug("Start computing interaction table.")
        table = NearFieldInteractionTable(
            dim=dim,
            quad_order=q_order,
            dtype=self.dtype,
            kernel_func=knl_func,
            kernel_type=knl_type,
            sumpy_kernel=sumpy_knl,
            build_method=compute_method,
            source_box_extent=self.root_extent * (2 ** (-source_box_level)),
            **self.table_extra_kwargs
        )

        if 0:
            # self-similarly shrink delta
            if 'delta' in kwargs:
                delta = kwargs.pop('delta') * (2 ** (-source_box_level))
                kwargs['delta'] = delta

        table.build_table(cl_ctx, queue, **kwargs)
        assert table.is_built

        # update database
        logger.debug("Start updating database.")

        grp = self.datafile[str(dim) + "D"][kernel_type]["Order_" + str(q_order)][
            "Level_" + str(source_box_level)
        ]

        grp.attrs["n_q_points"] = table.n_q_points
        grp.attrs["quad_order"] = table.quad_order
        grp.attrs["dim"] = table.dim
        grp.attrs["n_pairs"] = table.n_pairs
        grp.attrs["source_box_level"] = source_box_level
        grp.attrs["source_box_extent"] = self.root_extent * (
                2 ** (-source_box_level))

        for key, kval in kwargs.items():
            if isinstance(kval, (int, float, complex, str)):
                grp.attrs[key] = kval

        self.update_dataset(grp, "q_points", table.q_points)
        self.update_dataset(grp, "data", table.data)
        self.update_dataset(grp, "mode_normalizers", table.mode_normalizers)
        self.update_dataset(grp, "kernel_exterior_normalizers",
                table.kernel_exterior_normalizers)
        self.update_dataset(grp, "interaction_case_vecs",
                table.interaction_case_vecs)
        self.update_dataset(grp, "case_indices", table.case_indices)

        distinct_numbers = set()
        for vec in table.interaction_case_vecs:
            for l in vec:
                distinct_numbers.add(l)
        base = len(range(min(distinct_numbers), max(distinct_numbers) + 1))
        shift = -min(distinct_numbers)

        grp.attrs["case_encoding_base"] = base
        grp.attrs["case_encoding_shift"] = shift

        return table


# }}} End table dataset manager class

# vim: filetype=pyopencl:foldmethod=marker
